from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport.Plugins.Renderer import Renderer
from CGATReport.Plugins.Plotter import parseRanges
from CGATReport import Stats
from CGATReport.DataTree import path2str

from docutils.parsers.rst import directives

import re
try:
    import bokeh.plotting as bk
    HAS_BOKEH = True
except ImportError:
    HAS_BOKEH = False


class BokehPlotter():

    """Base class for Renderers that do simple 2D plotting.

    This mixin class provides convenience function
    for:class:`Renderer.Renderer` classes that do 2D plotting.

    The base class takes care of counting the plots created, provided
    that the methods:meth:`startPlot` and:meth:`endPlot` are called
    appropriately. It then inserts the appropriate place holders.

    This class adds the following options to the:term:`report` directive:

    :term:`logscale`: apply logscales one or more axes.

    :term:`xtitle`: add a label to the X axis

    :term:`ytitle`: add a label to the Y axis

    :term:`title`:  title of the plot

    :term:`add-title`: add title to each plot

    :term:`legend-location`: specify the location of the legend

    :term:`xrange`: restrict plot a part of the x-axis

    :term:`yrange`: restrict plot a part of the y-axis

    :term:`function`: add a function to the plot. Multiple
          functions can be supplied as a ,-separated list.

    :term:`vline`: add one or more vertical lines to the
          plot.

    :term:`xformat`: label for X axis

    :term:`yformat`: label for Y axis

    With some plots default layout options will result in plots
    that are misaligned (legends truncated, etc.). To fix this it might
    be necessary to increase plot size, reduce font size, or others.
    The following options will be passed on the matplotlib to permit
    this control.

    :term:`mpl-figure`: options for matplotlib
           ``figure`` calls().

    :term:`mpl-legend`: options for matplotlib
           ``legend`` calls().

    :term:`mpl-subplot`: options for matplotlib
           ``subplots_adjust`` calls().

    :term:`mpl-rc`: general environment settings for matplotlib.
          See the matplotlib documentation. Multiple options can be
          separated by ;, for example
          ``:mpl-rc: figure.figsize=(20,10);legend.fontsize=4``

    """

    mLegendFontSize = 8
    # number of chars to use to reduce legend font size
    mMaxLegendSize = 100

    # maximum number of rows per column. If there are more,
    # the legend is split into multiple columns
    mLegendMaxRowsPerColumn = 30

    options = (
        ('logscale',  directives.unchanged),
        ('title',  directives.unchanged),
        ('add-title',  directives.flag),
        ('xtitle',  directives.unchanged),
        ('ytitle',  directives.unchanged),
        ('xrange',  directives.unchanged),
        ('yrange',  directives.unchanged),
        ('zrange', directives.unchanged),
        ('function', directives.unchanged),
        ('vline', directives.unchanged),
        ('legend-location',  directives.unchanged),
        ('xformat', directives.unchanged),
        ('yformat', directives.unchanged),
    )

    format_colors = "bgrcmk"
    format_markers = "so^>dph8+x"
    format_lines = ('-', ':', '--')
    mPatterns = [None, '/', '\\', '|', '-', '+', 'x', 'o', 'O', '.', '*']

    def __init__(self, *args, **kwargs):
        """parse option arguments."""

        self.mFigure = 0

        self.logscale = kwargs.get("logscale", None)
        self.title = kwargs.get("title", None)
        self.add_title = "add-title" in kwargs
        self.xlabel = kwargs.get("xtitle", None)
        self.ylabel = kwargs.get("ytitle", None)
        self.functions = kwargs.get("function", None)
        self.vline = kwargs.get("vline", None)
        self.tight_layout = 'tight' in kwargs

        if self.functions:
            if "," in self.functions:
                self.functions = self.functions.split(",")
            else:
                self.functions = [self.functions]

        # substitute '-' in CGATReport-speak for ' ' in matplotlib speak
        self.legend_location = re.sub(
            "-", " ", kwargs.get("legend-location", "upper right"))
        # ("outer-top"))
        self.xrange = parseRanges(kwargs.get("xrange", None))
        self.yrange = parseRanges(kwargs.get("yrange", None))
        self.zrange = parseRanges(kwargs.get("zrange", None))

        self.xformat = kwargs.get("xformat", None)
        self.yformat = kwargs.get("yformat", None)

    def startPlot(self, **kwargs):
        """prepare everything for a plot.

        returns the current figure.
        """
        bk.figure()
        bk.output_file('/dev/null')
        bk.hold()

    def endPlot(self, plts, legends, path):
        """close plots.
        """

        title = path2str(path)
        figid = 10
        lines = []
        plot = self.plots[-1]
        figid = plot._id
        lines.append("")
        lines.append("#$bkh %s$#" % figid)
        lines.append("")
        r = ResultBlock("\n".join(lines), title=title)
        r.bokeh = self.plots[-1]
        return ResultBlocks(r)


class LinePlot(Renderer, BokehPlotter):

    '''create a line plot.

    This:class:`Renderer` requires at least three levels with

    line / label / coords

    This is a base class that provides several hooks for
    derived classes.

    initPlot()

    for line, data in work:
        initLine()

        for label, coords in data:
            xlabel, ylabels = initCoords()
            for ylabel in ylabels:
                addData(xlabel, ylabel)
            finishCoords()
        finishLine()

    finishPlot()

    This plotter accepts the following options:

    :term:`as-lines`: do not plot symbols
    :term:`yerror`: every second data track is a y error

    '''
    nlevels = 2

    options = BokehPlotter.options +\
        (('as-lines', directives.flag),
         ('yerror', directives.flag),
         )

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        BokehPlotter.__init__(self, *args, **kwargs)

        self.as_lines = "as-lines" in kwargs

        # data to use for Y error bars
        self.yerror = "yerror" in kwargs

        # do not plot more than ten tracks in one plot
        self.split_at = 10

        self.format_colors = bk.brewer["Spectral"][10]

    def initPlot(self, fig, dataseries, path):
        '''initialize plot.'''

        self.legend = []
        self.plots = []
        self.xlabels = []
        self.ylabels = []

    def getFormat(self, nplotted):
        '''return tuple with color, linestyle and marker for
        data series *n* within a plot.'''

        color = self.format_colors[nplotted % len(self.format_colors)]
        nplotted /= len(self.format_colors)
        linestyle = self.format_lines[nplotted % len(self.format_lines)]
        if self.as_lines:
            marker = None
        else:
            nplotted /= len(self.format_lines)
            marker = self.format_markers[nplotted % len(self.format_markers)]

        return color, linestyle, marker

    def addData(self,
                xvals, yvals,
                xlabel, ylabel,
                nplotted,
                yerrors=None):

        xxvals, yyvals = Stats.filterNone((xvals, yvals))

        color, linestyle, marker = self.getFormat(nplotted)

        self.plots.append(bk.line(
            xxvals,
            yyvals,
            color=color,
            line_width=2))
        # other options:
        # title=...
        # legend=...

    def initLine(self, line, data):
        '''hook for code working on a line.'''
        pass

    def initCoords(self, label, coords):
        '''hook for code working a collection of coords.

        should return a single key for xvalues and
        one or more keys for y-values.
        '''
        keys = list(coords.keys())
        return keys[0], keys[1:]

    def finishCoords(self, label, coords):
        '''hook called after all coords have been processed.'''
        pass

    def finishPlot(self, fig, work, path):
        '''hook called after plotting has finished.'''
        # plt.xlabel("-".join(set(self.xlabels)))
        # plt.ylabel("-".join(set(self.ylabels)))

    def render(self, dataframe, path):

        fig = self.startPlot()

        paths = dataframe.index.get_level_values(0).unique()

        self.initPlot(fig, dataframe, path)

        nplotted = 0

        columns = dataframe.columns

        if len(columns) < 2:
            raise ValueError(
                'require at least two columns, got %i' % len(columns))

        for ppath in paths:
            work = dataframe.ix[ppath]
            xvalues = work[columns[0]]
            for column in work.columns[1:]:

                self.initLine(column, work)

                yvalues = work[column]

                if len(xvalues) != len(yvalues):
                    raise ValueError("length of x,y tuples not consistent: "
                                     "(%s) %i != (%s) %i" %
                                     (columns[0],
                                      len(xvalues),
                                      column,
                                      len(yvalues)))

                self.initCoords(xvalues, yvalues)

                self.addData(xvalues,
                             yvalues,
                             columns[0],
                             column,
                             nplotted)
                nplotted += 1

                # self.legend.append(path2str(ppath) + "/" + column)

        self.finishPlot(fig, dataframe, path)

        return self.endPlot(self.plots, self.legend, path)



__all__ = [
    "CorrespondenceAnalysis",
    "Stats",
    "DataTypes",
    "Logger",
    "Component",
    "Utils",
    "Cache",
    "ResultBlock",
    "Dispatcher",
    "DataTree",
    # Trackers to export
    "Tracker",
    "Config",
    "Empty",
    "Status",
    # directives to export
    "only_directive",
    "report_directive"]



import copy


class DataSimple(object):

    """Base class for data types.

    Derived classes enforce consistency checks on data.
    """
    __slots__ = ["_instance", "_data", "_fn"]

    def __init__(self, fn):
        """Store data returned by function."""
        object.__setattr__(self, "_fn", fn)
        object.__setattr__(self, "_data", None)
        object.__setattr__(self, "_instance", None)

    def __call__(self, *args, **kwargs):
        """call the function and return a clone of one-self.
        """
        # Decorators will use the same object for each decoration
        # and data will get overwritten in successive calls to the same function.
        # Thus clone oneself before storing the data and return
        # the clone.
        clone = copy.copy(self)
        setattr(clone, "_data", self._fn(self._instance, *args, **kwargs))
        clone.__check__()
        return clone

    def __len__(self):
        return len(self._data)

    def __get__(self, instance, cls=None):
        object.__setattr__(self, "_instance", instance)
        return self

    def __getstate__(self):
        # previously used deepcoy, but not necessary
        return {"_data": self._data}

    def __setstate__(self, dict):
        for key, val in list(dict.items()):
            object.__setattr__(self, key, val)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, *args, **kwargs):
        return self._data.__getitem__(*args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        return self._data.__getslice__(*args, **kwargs)

    def __contains__(self):
        return self._data.__contains__(*args, **kwargs)

    def __copy__(self):
        return self.__class__(self)
#    def __getattr__(self, name):
#        return getattr(self._data, name)
#    def __setattr__(self, name, value):
#        setattr(self._data, name, value)


class Data(object):

    """Base class for data types.

    Derived classes enforce consistency checks on data.
    """
    __slots__ = ["_data"]

    def __init__(self, data):
        """Store data returned by function."""
        object.__setattr__(self, "_data", data)
        if data:
            self.__check__()

    def __len__(self):
        return len(self._data)

    def __getstate__(self):
        # previously used deepcoy, but not necessary
        return {"_data": self._data}

    def __setstate__(self, dict):
        for key, val in list(dict.items()):
            object.__setattr__(self, key, val)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, *args, **kwargs):
        return self._data.__getitem__(*args, **kwargs)

    def __setslice__(self, *args, **kwargs):
        return self._data.__getslice__(*args, **kwargs)

    def __contains__(self):
        return self._data.__contains__(*args, **kwargs)

    def __copy__(self):
        return self.__class__(self)
#    def __getattr__(self, name):
#        return getattr(self._data, name)
#    def __setattr__(self, name, value):
#        setattr(self._data, name, value)


class SingleColumn(Data):

    """Single column.

    The data can be any scalar type.

    Example: (1,2,"a")
    """

    def __init__(self, fn):
        Data.__init__(self, fn)

    def __check__(self):
        assert type(self._data) in ContainerTypes, "returned type is not a collection: %s" % (
            type(self._data))
        for x in self._data:
            assert type(x) in NumberTypes, "value %s is not a number: type=%s" % (
                str(x), type(x))


class SingleColumnData(Data):

    """Single column data.

    All data are numerical values.

    Example: (1,2,3)
    """

    def __init__(self, fn):
        Data.__init__(self, fn)

    def __check__(self):
        assert type(self._data) in ContainerTypes, "returned type is not a collection: %s" % (
            type(self._data))
        for x in self._data:
            assert type(x) in NumberTypes, "value %s is not a number: type=%s" % (
                str(x), type(x))


class MultipleColumns(Data):

    """Multiple column data

    The data can be any scalar type. All columns have the same length.

    Example: (("column1", "column2"), (("val1",2,3), ("val2",2,3)))
    """

    def __init__(self, fn):
        Data.__init__(self, fn)

    def __check__(self):
        assert type(self._data) in ContainerTypes, "returned type is not a collection: %s" % (
            type(self._data))
        assert type(self._data[0]) in ContainerTypes, "first column is not a collection: %s" % (
            type(self._data[0]))
        assert type(self._data[1]) in ContainerTypes, "second column is not a collection: %s" % (
            type(self._data[1]))
        for c in self._data[1]:
            assert type(
                c) in ContainerTypes, "column is not a collection: %s" % (type(c))
        try:
            assert min([len(c) for c in self._data[1]]) == max([len(c) for c in self._data[1]]), \
                "data columns have not the same length: %i != %i." %\
                (min([len(c) for c in self._data[1]]), max([len(c)
                                                            for c in self._data[1]]))
        except ValueError as msg:
            # ignore errors due to empty sequences
            pass


class MultipleColumnData(Data):

    """Multiple column data

    All data are numerical values.

    Example: (("column1", "column2"), ((1,2,3), (1,2,3)))
    """

    def __init__(self, fn):
        Data.__init__(self, fn)

    def __check__(self):
        assert type(self._data) in ContainerTypes, "returned type is not a collection: %s" % (
            type(self._data))
        assert type(self._data[0]) in ContainerTypes, "first field is not a collection: %s" % (
            type(self._data[0]))
        assert type(self._data[1]) in ContainerTypes, "second field is not a collection: %s" % (
            type(self._data[1]))
        for c in self._data[1]:
            assert type(
                c) in ContainerTypes, "column is not a collection: %s" % (type(c))
            for x in c:
                assert type(x) in NumberTypes, "value %s is not a number: type=%s" % (
                    str(x), type(x))
        assert min([len(c) for c in self._data[1]]) == max(
            [len(c) for c in self._data[1]]), "data columns have not the same length."


class LabeledData(Data):

    """Labeled data points.

    Data can be of any type.  There is only one value per label.

    Example: (("column1", 1), ("column2",2))
    """

    def __init__(self, fn):
        Data.__init__(self, fn)

    def __check__(self):
        assert type(
            self._data) in ContainerTypes, "returned type is not a collection: %s" % (self._data)
        for x in self._data:
            assert type(
                x) in ContainerTypes, "row is not a collection: %s" % str(x)
            assert len(
                x) == 2, "data is not a column, value tuple: %s" % str(x)


def returnLabeledValue(Data):
    """decorator for Trackers returning:class:`LabeledValue`."""
    def wrapped_f(*args, **kwargs):
        return LabeledValue(f(*args, **kwargs))
    return wrapped_f


def returnSingleColumn(f):
    """decorator for Trackers returning:class:`SingleColumn`."""
    def wrapped_f(*args, **kwargs):
        return SingleColumn(f(*args, **kwargs))
    return wrapped_f


def returnSingleColumnData(f):
    """decorator for Trackers returning:class:`SingleColumnData`."""
    def wrapped_f(*args, **kwargs):
        return SingleColumnData(f(*args, **kwargs))
    return wrapped_f


def returnMultipleColumns(f):
    """decorator for Trackers returning:class:`MultipleColumn`."""
    def wrapped_f(*args, **kwargs):
        return MultipleColumns(f(*args, **kwargs))
    return wrapped_f


def returnMultipleColumnData(f):
    """decorator for Trackers returning:class:`MultipleColumnData`."""
    def wrapped_f(*args, **kwargs):
        return MultipleColumnData(f(*args, **kwargs))
    return wrapped_f


def returnLabeledData(f):
    """decorator for Trackers returning:class:`LabeledData`."""
    def wrapped_f(*args, **kwargs):
        return LabeledData(f(*args, **kwargs))
    return wrapped_f

# def returnSingleColumn(f):
#     """decorator for Trackers returning:class:`SingleColumn`."""
#     return SingleColumn(f)

# def returnSingleColumnData(f):
#     """decorator for Trackers returning:class:`SingleColumnData`."""
#     return SingleColumnData(f)

# def returnMultipleColumns(f):
#     """decorator for Trackers returning:class:`MultipleColumn`."""
#     return MultipleColumns(f)

# def returnMultipleColumnData(f):
#     """decorator for Trackers returning:class:`MultipleColumnData`."""
#     return MultipleColumnData(f)

# def returnLabeledData(f):
#     """decorator for Trackers returning:class:`LabeledData`."""
#     return LabeledData(f)



from CGATReport.Tracker import Tracker

import re
import os
import six


class WordCounterWithSlices(Tracker):
    """Counting word size."""

    def getTracks(self):
        return ("all", ".py", ".rst")

    def getSlices(self):
        return ("any", "vocals", "consonants")

    def __call__(self, track, slice):
        word_sizes = []

        if track == "all":
            tracks = [".py", ".rst"]
        else:
            tracks = [track]

        if slice == "any":
            test_f = lambda x: True
        elif slice == "vocals":
            test_f = lambda x: x[0].upper() in "AEIOU"
        elif slice == "consonants":
            test_f = lambda x: x[0].upper() not in "BCDFGHJKLMNPQRSTVWXYZ"

        for root, dirs, files in os.walk(os.path.dirname(__file__)):
            for f in files:
                fn, ext = os.path.splitext(f)
                if ext not in tracks:
                    continue
                if six.PY2:
                    infile = open(os.path.join(root, f), "r")
                else:
                    infile = open(os.path.join(root, f), "r", encoding="utf-8")
                words = [
                    w for w in
                    re.split("\s+", "".join(infile.readlines()))
                    if len(w) > 0]
                word_sizes.extend([len(w) for w in words if test_f(w)])
                infile.close()

        return {"word sizes": word_sizes}



import unittest
import subprocess
import collections
import os
import re
import tempfile
import shutil
import sys

N_CORES = 4


class TestReportBuilding(unittest.TestCase):

    ignored_errors = [
        "testing",
        "UnknownTracker",
        "GalleryBokeh",
        "TestExceptions",
        "unknown-renderer",
        "unknown-transform"]

    def setUp(self):
        # build_dir = tempfile.mkdtemp(prefix="test_report-results.dir.",
        #                             dir=os.path.abspath(os.curdir))


        self.build_dir = os.path.join(os.path.abspath(os.curdir),
                                      "test_report-results-{}.dir".format(
                                          sys.version.split(" ")[0]))

        self.docs_dir = os.path.join(
            os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
            "doc")
        
        if not os.path.exists(self.docs_dir):
            raise ValueError("doc directory {} does not exist".format(self.docs_dir))

        print ("docs_dir is {}".format(self.docs_dir))
        print ("build_dir is {}".format(self.build_dir))

    def parse_cgatreport_log(self):
        
        errors = []
        counter = collections.defaultdict(int)
        with open(os.path.join(self.build_dir, "cgatreport.log")) as inf:
            for line in inf:
                parts = line[:-1].split(" ")
                if len(parts) < 3:
                    continue
                # ignore errors created in test suite
                skip = False
                for pattern in self.ignored_errors:
                    if pattern in line:
                        skip = True
                        break
                if skip:
                    continue
                counter[parts[2]] += 1
                if parts[2] == "ERROR":
                    errors.append(line)

        # shutil.rmtree(build_dir)
        # filter version specific errors
        if sys.version_info.major == 3 and sys.version_info.minor >= 5:
            # statsmodels 0.6.1 is not py3.5 compatible
            errors = [x for x in errors if not re.search(
                    "seaborn raised error for statement"
                    ".*"
                    "slice indices must be "
                    "integers or None or have an __index__ method", x)]
            errors = [x for x in errors if not re.search(
                    "exception in rendering: <class 'TypeError'>"
                    ".*"
                    "slice indices must be "
                    "integers or None or have an __index__ method", x)]

            counter["ERROR"] = len(errors)
        return counter, errors

    def test_full_docs_build_without_error(self):

        # "cgatreport-build --num-jobs={n_cores} "

        subprocess.check_output(
            "rm -rf {build_dir} && "
            "mkdir {build_dir} && "
            "gunzip < {docs_dir}/csvdb_data.txt.gz | sqlite3 {build_dir}/csvdb && "
            "cp -r {docs_dir}/images {build_dir} && "
            "cp {docs_dir}/*.ini {build_dir}/ && "
            "cd {build_dir} && "
            "xvfb-run -d "
            "sphinx-build "
            "-j {n_cores} "
            "-b html "
            "-d _build/doctrees "
            "{docs_dir} "
            "_build/html >& {build_dir}/build.log".format(
                build_dir=self.build_dir,
                docs_dir=self.docs_dir,
                n_cores=N_CORES),
            shell=True)
        
        counter, errors = self.parse_cgatreport_log()

        self.assertEqual(
            counter["ERROR"], 0,
            "List of errors:\n{lines}\n"
            "There were {nerrors} errors during report building. "
            "See {build_dir}/cgatreport.log "
            "for more information".format(
                lines="".join(errors),
                build_dir=self.build_dir,
                nerrors=counter["ERROR"]))

    def test_quickstart_report_will_pass(self):

        subprocess.check_output(
            "rm -rf {build_dir} && "
            "mkdir {build_dir} && "
            "cgatreport-quickstart -d {build_dir} >& {build_dir}/quickstart.log && "
            "make -C {build_dir} html >& {build_dir}/build.log".format(
                build_dir=self.build_dir,
                docs_dir=self.docs_dir,
                n_cores=N_CORES),
            shell=True)

        counter, errors = self.parse_cgatreport_log()

        self.assertEqual(
            counter["ERROR"], 0,
            "List of errors:\n{lines}\n"
            "There were {nerrors} errors during report building. "
            "See {build_dir}/cgatreport.log "
            "for more information".format(
                lines="".join(errors),
                build_dir=self.build_dir,
                nerrors=counter["ERROR"]))


if __name__ == "__main__":
    unittest.main()    



"""set and manipulate options for Capabilities.
"""

from docutils.parsers.rst import directives

from CGATReport.Capabilities import get_all_plugins
from CGATReport.Utils import get_params


OPTIONS = None


def get_option_map():

    global OPTIONS

    if OPTIONS is None:
        OPTIONS = {}
        for section, plugins in get_all_plugins().items():
            OPTIONS[section] = {}
            for name, cls in list(plugins.items()):
                try:
                    OPTIONS[section].update(dict(cls.options))
                except AttributeError:
                    pass

        OPTIONS["dispatch"] = {
            'groupby': directives.unchanged,
            'set-index': directives.unchanged,
            'include-columns': directives.unchanged,
            'exclude-columns': directives.unchanged,
            'tracks': directives.unchanged,
            'slices': directives.unchanged,
            'layout': directives.unchanged,
            'restrict': directives.unchanged,
            'exclude': directives.unchanged,
            'nocache': directives.flag,
        }

        # options used in trackers
        OPTIONS["tracker"] = {
            'regex': directives.unchanged,
            'glob': directives.unchanged,
            'sql_backend': directives.unchanged,
            'tracker': directives.unchanged,
        }

        OPTIONS["display"] = {
            # general image options
            'alt': directives.unchanged,
            'height': directives.length_or_unitless,
            'width': directives.length_or_percentage_or_unitless,
            'scale': directives.nonnegative_int,
            'class': directives.class_option,
            # 'align': align,
            # options for cgatreport
            'render': directives.unchanged,
            'transform': directives.unchanged,
            'display': directives.unchanged,
            'extra-formats': directives.unchanged,
            'no-caption': directives.flag,
            'no-title': directives.flag,
            'no-links': directives.flag,
        }

    return OPTIONS


def get_option_spec():
    '''build option spec for sphinx

    This method returns a flattened :var:`options`
    dictionary.
    '''
    r = {}
    for x, xx in get_option_map().items():
        r.update(xx)

    # add the primary actor options
    r["render"] = directives.unchanged
    r["transform"] = directives.unchanged
    return r


def update_options(kwargs):
    '''replace placeholders in kwargs with with PARAMS read from config file.

    returns the update dictionary.
    '''

    params = get_params()

    for key, value in list(kwargs.items()):
        try:
            v = value.strip()
        except AttributeError:
            # ignore non-string types
            continue

        if v.startswith("@") and v.endswith("@"):
            code = v[1:-1]
            if code in params:
                kwargs[key] = params[code]
            else:
                raise ValueError("unknown placeholder `%s`" % code)

    return kwargs


def expand_option(option):
    """expand option that is a ',' separated list of assignments.
    """
    kwargs = {}
    option = option.strip()
    if len(option) == 0:
        return kwargs
    parts = option.split(",")

    for part in parts:
        if not part.strip():
            continue
        if "=" not in part:
            raise ValueError("malformed tracker options '{}' in '{}'".format(part, option))
        key, val = part.split("=", 1)
        try:
            kwargs[key.strip()] = eval(val.strip())
        except NameError:
            kwargs[key.strip()] = val

    return kwargs


def select_and_delete_options(options, select, expand=[]):
    '''collect options in *select* and from *options* and remove those found.

    expand is a list of keywards that will be expanded.

    returns dictionary of options found.
    '''
    new_options = {}
    for k, v in options.items():
        if k in select:
            new_options[k] = v
    for k in new_options.keys():
        del options[k]

    for k in expand:
        if k in new_options:
            new_options.update(expand_option(new_options[k]))
    return new_options



import imp
import io
import os
import re
import six
import sys
import types

from logging import debug, warning, critical
from CGATReport.Utils import memoized
from CGATReport.Types import get_encoding

import CGATReport.Plugins
import CGATReport.Plugins.BokehPlotter
import CGATReport.Plugins.BokehPlugin
import CGATReport.Plugins.GGPlotter
import CGATReport.Plugins.HTMLPlugin
import CGATReport.Plugins.MatplotlibPlugin
import CGATReport.Plugins.PandasPlotter
import CGATReport.Plugins.Plotter
import CGATReport.Plugins.RPlotPlugin
import CGATReport.Plugins.RPlotter
import CGATReport.Plugins.RSTPlugin
import CGATReport.Plugins.Renderer
import CGATReport.Plugins.Seaborn
import CGATReport.Plugins.SlideShow
import CGATReport.Plugins.Transformer
import CGATReport.Plugins.TransformersGeneLists
import CGATReport.Plugins.XLSPlugin

CapabilityMap = {
    "collect":
    {
        "matplotlib":
        CGATReport.Plugins.MatplotlibPlugin.MatplotlibPlugin,
        "rplot":
        CGATReport.Plugins.RPlotPlugin.RPlotPlugin,
        "html":
        CGATReport.Plugins.HTMLPlugin.HTMLPlugin,
        "rst":
        CGATReport.Plugins.RSTPlugin.RSTPlugin,
        "xls":
        CGATReport.Plugins.XLSPlugin.XLSPlugin,
        "bokeh":
        CGATReport.Plugins.BokehPlugin.BokehPlugin,
    },
    "transform":
    {
        "stats":
        CGATReport.Plugins.Transformer.TransformerStats,
        "correlation":
        CGATReport.Plugins.Transformer.TransformerCorrelationPearson,
        "pearson":
        CGATReport.Plugins.Transformer.TransformerCorrelationPearson,
        "contingency":
        CGATReport.Plugins.Transformer.TransformerContingency,
        "spearman":
        CGATReport.Plugins.Transformer.TransformerCorrelationSpearman,
        "test-mwu":
        CGATReport.Plugins.Transformer.TransformerMannWhitneyU,
        "aggregate":
        CGATReport.Plugins.Transformer.TransformerAggregate,
        "histogram":
        CGATReport.Plugins.Transformer.TransformerHistogram,
        "histogram-stats":
        CGATReport.Plugins.Transformer.TransformerHistogramStats,
        "filter":
        CGATReport.Plugins.Transformer.TransformerFilter,
        "pandas":
        CGATReport.Plugins.Transformer.TransformerPandas,
        "melt":
        CGATReport.Plugins.Transformer.TransformerMelt,
        "pivot":
        CGATReport.Plugins.Transformer.TransformerPivot,
        "hypergeometric":
        CGATReport.Plugins.TransformersGeneLists.TransformerHypergeometric,
        "venn":
        CGATReport.Plugins.TransformersGeneLists.TransformerVenn,
        "p-adjust":
        CGATReport.Plugins.TransformersGeneLists.TransformerMultiTest,
        "odds-ratio":
        CGATReport.Plugins.TransformersGeneLists.TransformerOddsRatio,
        },
    "render":
    {
        "user":
        CGATReport.Plugins.Renderer.User,
        "debug":
        CGATReport.Plugins.Renderer.Debug,
        "dataframe":
        CGATReport.Plugins.Renderer.DataFrame,
        "table":
        CGATReport.Plugins.Renderer.Table,
        "rst-table":
        CGATReport.Plugins.Renderer.RstTable,
        "xls-table":
        CGATReport.Plugins.Renderer.XlsTable,
        "html-table":
        CGATReport.Plugins.Renderer.HTMLTable,
        "glossary-table":
        CGATReport.Plugins.Renderer.GlossaryTable,
        "matrix":
        CGATReport.Plugins.Renderer.TableMatrix,
        "matrixNP":
        CGATReport.Plugins.Renderer.NumpyMatrix,
        "status":
        CGATReport.Plugins.Renderer.Status,
        "status-matrix":
        CGATReport.Plugins.Renderer.StatusMatrix,
        "line-plot":
        CGATReport.Plugins.Plotter.LinePlot,
        # for backwards compatibility
        "density-plot":
        CGATReport.Plugins.Seaborn.KdePlot,
        "histogram-plot":
        CGATReport.Plugins.Plotter.HistogramPlot,
        "pie-plot":
        CGATReport.Plugins.Plotter.PiePlot,
        "scatter-plot":
        CGATReport.Plugins.Plotter.ScatterPlot,
        "scatter-rainbow-plot":
        CGATReport.Plugins.Plotter.ScatterPlotWithColor,
        "matrix-plot":
        CGATReport.Plugins.Plotter.TableMatrixPlot,
        "matrixNP-plot":
        CGATReport.Plugins.Plotter.NumpyMatrixPlot,
        "hinton-plot":
        CGATReport.Plugins.Plotter.HintonPlot,
        "gallery-plot":
        CGATReport.Plugins.Plotter.GalleryPlot,
        "slideshow-plot":
        CGATReport.Plugins.SlideShow.SlideShowPlot,
        "bar-plot":
        CGATReport.Plugins.Plotter.BarPlot,
        "stacked-bar-plot":
        CGATReport.Plugins.Plotter.StackedBarPlot,
        "interleaved-bar-plot":
        CGATReport.Plugins.Plotter.InterleavedBarPlot,
        "venn-plot":
        CGATReport.Plugins.Plotter.VennPlot,
        # ggplot
        "ggplot":
        CGATReport.Plugins.GGPlotter.GGPlot,
        # pandas plotting
        "pdplot":
        CGATReport.Plugins.PandasPlotter.PandasPlot,
        # seaborn plots
        "sb-box-plot":
        CGATReport.Plugins.Seaborn.BoxPlot,
        "sb-violin-plot":
        CGATReport.Plugins.Seaborn.ViolinPlot,
        "sb-kde-plot":
        CGATReport.Plugins.Seaborn.KdePlot,
        "sb-pair-plot":
        CGATReport.Plugins.Seaborn.PairPlot,
        "sb-dist-plot":
        CGATReport.Plugins.Seaborn.DistPlot,
        "sb-heatmap-plot":
        CGATReport.Plugins.Seaborn.HeatmapPlot,
        "sb-clustermap-plot":
        CGATReport.Plugins.Seaborn.ClustermapPlot,
        # R plots
        "r-line-plot":
        CGATReport.Plugins.RPlotter.LinePlot,
        "r-box-plot":
        CGATReport.Plugins.RPlotter.BoxPlot,
        "r-smooth-scatter-plot":
        CGATReport.Plugins.RPlotter.SmoothScatterPlot,
        "r-heatmap-plot":
        CGATReport.Plugins.RPlotter.HeatmapPlot,
        "r-ggplot":
        CGATReport.Plugins.RPlotter.GGPlot,
        # Bokeh plots
        "b-line-plot":
        CGATReport.Plugins.BokehPlotter.LinePlot,
        # backwards compatibility
        "box-plot":
        CGATReport.Plugins.Seaborn.BoxPlot,
        "violin-plot":
        CGATReport.Plugins.Seaborn.ViolinPlot
    }
}


@memoized
def get_module(name):
    """load module in fullpat
    """
    # remove leading '.'
    debug("entered get_module with `%s`" % name)

    parts = name.split(".")
    if parts[0] == "Tracker":
        # special case: Trackers shipped with CGATReport
        if len(parts) > 2:
            raise NotImplementedError("built-in trackers are Tracker.<name> ")
        name = "Tracker"
        path = [os.path.join(CGATReport.__path__[0])]

    # the first part needs to be on the python sys.path
    elif len(parts) > 1:
        try:
            (file, pathname, description) = imp.find_module(parts[0])
        except ImportError as msg:
            warning("could not find module %s: msg=%s" % (name, msg))
            raise ImportError("could not find module %s: msg=%s" % (name, msg))

        path = [os.path.join(pathname, *parts[1:-1])]
        name = parts[-1]
    else:
        path = None

    debug("searching for module name=%s at path=%s" % (name, str(path)))

    # find module
    try:
        (modulefile, pathname, description) = imp.find_module(name, path)
    except ImportError as msg:
        warning("could not find module %s in %s: msg=%s" % (name, path, msg))
        raise ImportError(
            "could not find module %s in %s: msg=%s" % (name, path, msg))

    if modulefile is None:
        warning("could not find module %s in %s" % (name, path))
        raise ImportError(
            "find_module returned None for %s in %s" %
            (name, path))

    stdout = sys.stdout
    sys.stdout = io.StringIO()
    debug("loading module: %s: %s, %s, %s" %
          (name, modulefile, pathname, description))
    # imp.load_module modifies sys.path - save original and restore
    oldpath = sys.path

    # add to sys.path to ensure that imports in the directory work
    if pathname not in sys.path:
        sys.path.append(os.path.dirname(pathname))

    try:
        module = imp.load_module(name, modulefile, pathname, description)
    except:
        warning("could not load module %s" % name)
        raise
    finally:
        modulefile.close()
        sys.stdout = stdout
        sys.path = oldpath

    return module, pathname


def get_code(cls, pathname):
    '''retrieve code for methods and functions.'''
    # extract code
    code = []
    if six.PY2:
        infile = open(pathname, "r")
    else:
        infile = open(pathname, "r", encoding=get_encoding())

    for line in infile:
        x = re.search("(\s*)(class|def)\s+%s" % cls, line)
        if x:
            indent = len(x.groups()[0])
            code.append(line)
            break
    for line in infile:
        if len(re.match("^(\s*)", line).groups()[0]) <= indent:
            break
        code.append(line)
    infile.close()

    return code


def is_class(obj):
    '''return true if obj is a class.

    return False if it is a function object.
    raise ValueError if neither
    '''

    # checking for subclass of Tracker causes a problem
    # for classes defined in the module Tracker itself.
    # The problem is that 'Tracker' is CGATReport.Tracker.Tracker,
    # while the one in tracker is simply Tracker
    # issubclass(obj, Tracker) and \

    if isinstance(obj, type) and \
            hasattr(obj, "__call__"):
        return True
    elif isinstance(obj, (type, types.FunctionType)):
        return False
    elif isinstance(obj, (type, types.LambdaType)):
        return False

    raise ValueError("can not make sense of tracker %s" % str(obj))


def make_object(path, args=(), kwargs={}):
    '''return object of type *path*

    This function is similar to an import statement, but
    also instantiates the class and returns the object.

    The object is instantiated with *args* and **kwargs**.
    '''

    # split class from module
    name, cls = os.path.splitext(path)

    # remove leading '.'
    cls = cls[1:]

    debug("instantiating class %s" % cls)

    module, pathname = get_module(name)

    # get class from module
    try:
        obj = getattr(module, cls)
    except AttributeError:
        raise AttributeError("module %s (%s) has no attribute '%s'" %
                             (module, pathname, cls))
    # instantiate, if it is a class
    if is_class(obj):
        try:
            obj = obj(*args, **kwargs)
        except AttributeError as msg:
            critical("instantiating class %s.%s failed: %s" %
                     (module, cls, msg))
            raise

    return obj, module, pathname, cls


@memoized
def make_tracker(path, args=(), kwargs={}):
    """retrieve an instantiated tracker and its associated code.

    returns a tuple (code, tracker, pathname).
    """
    obj, module, pathname, cls = make_object(path, args, kwargs)
    code = get_code(cls, pathname)
    return code, obj, pathname


@memoized
def make_renderer(path, args=(), kwargs={}):
    """retrieve an instantiated Renderer.

    returns the object.
    """
    obj, module, pathname, cls = make_object(path, args, kwargs)
    return obj


@memoized
def make_transformer(path, args=(), kwargs={}):
    """retrieve an instantiated Transformer.

    returns the object.
    """
    obj, module, pathname, cls = make_object(path, args, kwargs)
    return obj


def get_transformers(transformers, kwargs={}):
    '''find and instantiate all transformers.'''

    result = []
    for transformer in transformers:
        tt = CapabilityMap["transform"].get(transformer, None)
        if tt is not None:
            instance = tt(**kwargs)
        else:
            instance = make_transformer(transformer, (), kwargs)

        if not instance:
            msg = "could not instantiate transformer '%s'. Available transformers:\n  %s" % \
                (transformer,
                 "\n  ".join(sorted(CapabilityMap["transform"].keys())))
            raise KeyError(msg)

        result.append(instance)

    return result


def get_renderer(renderer_name, kwargs={}):
    '''find and instantiate renderer.'''

    instance = None

    tt = CapabilityMap["render"].get(renderer_name, None)
    if tt is not None:
        instance = tt(**kwargs)
    else:
        instance = make_renderer(renderer_name, kwargs)

    if not instance:
        raise KeyError(
            "could not instantiate renderer '%s'. Available renderers:\n  %s" %
            (renderer_name,
             ",".join(
                 sorted(CapabilityMap["render"].keys()))))

    return instance


def get_plugins(capability):
    return CapabilityMap[capability]


def get_all_plugins():
    return CapabilityMap



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from CGATReport.Tracker import Tracker
from collections import OrderedDict as odict


class LabeledDataExampleUnicode(Tracker):
    slices = ("slüce1", "slüce2")
    tracks = ("träck1", "träck2", "träck3")

    def __call__(self, track, slice=None):

        v = int(track[-1]) * int(slice[-1])
        if slice == "slüce1":
            return odict((("column1", v),
                          ("column2", v * 2),))
        elif slice == "slüce2":
            return odict((("column1", v),
                          ("column2", v * 2),
                          ("column3", v * 3),))



from docutils.parsers.rst import directives

from .Plotter import TableMatrixPlot, DataSeriesPlot
import pandas
import seaborn
import numpy
from CGATReport.Plugins.Renderer import Renderer
from CGATReport.Plugins.Plotter import Plotter


class SeabornPlot(object):

    options = (
        ('kwargs', directives.unchanged),
    )

    kwargs = ""

    def __init__(self, *args, **kwargs):
        self.kwargs = kwargs.get("kwargs", "")

    def execute(self, statement, g, l):
        """execute seaborn statement. Inserts kwargs."""
        statement = statement.strip()
        assert statement.endswith(")")
        if self.kwargs:
            statement = statement[:-1] + ", %s)" % self.kwargs

        plot = None
        try:
            exec("plot = %s" % statement, g, l)
        except Exception as msg:
            raise Exception(
                "seaborn raised error for statement '%s': msg=%s" %
                (statement, msg))
        return plot

    def build_kwargs_dict():
        pass

class PairPlot(Renderer, Plotter, SeabornPlot):

    options = Renderer.options +\
              Plotter.options +\
              SeabornPlot.options

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def render(self, dataframe, path):

        statement = ("seaborn.pairplot(dataframe)")
        plot = self.execute(statement, globals(), locals())

        return self.endPlot([plot], [], path)


class BoxPlot(DataSeriesPlot, SeabornPlot):

    """Write a set of box plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    options = DataSeriesPlot.options +\
              SeabornPlot.options

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def plotData(self, data, melted):
        if melted:
            return [seaborn.boxplot(data.value,
                                    groupby=data.label)]
        else:
            return [seaborn.boxplot(data)]


class ViolinPlot(BoxPlot):

    """Write a set of violin plots.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    options = BoxPlot.options

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        BoxPlot.__init__(self, *args, **kwargs)

    def plotData(self, data, melted):
        if melted:
            return [seaborn.violinplot(data.value,
                                       groupby=data.label)]
        else:
            return [seaborn.violinplot(data)]


class KdePlot(DataSeriesPlot, SeabornPlot):

    """Write a set of density plots.

    The data series is converted to a kernel density
    estimate.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """
    options = DataSeriesPlot.options +\
              SeabornPlot.options

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def plotData(self, dataframe, melted):
        plts = []
        if melted:
            for key, group in dataframe.groupby(dataframe.label):
                plot = self.execute(
                    "seaborn.kdeplot( "
                    "numpy.array(dataframe.value, dtype=numpy.float), "
                    "label=key)", globals(), locals())
                plts.append(plot)
        else:
            for column in dataframe.columns:
                plot = self.execute(
                    "seaborn.kdeplot( "
                    "numpy.array(dataframe[column], dtype=numpy.float), "
                    "label=column)", globals(), locals())
        return plts


class DistPlot(DataSeriesPlot, SeabornPlot):
    """Write a set of density plots.

    The data series is converted to a kernel density
    estimate.

    This:class:`Renderer` requires two levels.

    labels[dict] / data[array]
    """

    options = DataSeriesPlot.options + SeabornPlot.options

    def __init__(self, *args, **kwargs):
        DataSeriesPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def plotData(self, dataframe, melted):
        plts = []
        if melted:
            for key, group in dataframe.groupby(dataframe.label):
                plot = self.execute(
                    "seaborn.distplot( "
                    "numpy.array(dataframe.value, dtype=numpy.float), "
                    "label=key)", globals(), locals())
                plts.append(plot)
        else:
            for column in dataframe.columns:
                plot = self.execute(
                    "seaborn.distplot( "
                    "numpy.array(dataframe[column], dtype=numpy.float), "
                    "label=column)", globals(), locals())
                plts.append(plot)
        return plts


class SeabornMatrixPlot(TableMatrixPlot, SeabornPlot):

    options = TableMatrixPlot.options + SeabornPlot.options

    def __init__(self, *args, **kwargs):
        TableMatrixPlot.__init__(self, *args, **kwargs)
        SeabornPlot.__init__(self, *args, **kwargs)

    def addColourBar(self):
        pass

    def startPlot(self, **kwargs):

        # seaborn will create figure, so do not create one.
        self.mFigure += 1

    def plotMatrix(self, matrix, row_headers, col_headers,
                   vmin, vmax, colorscheme=None):

        self.debug("plot matrix started")
        data = pandas.DataFrame(matrix,
                                columns=col_headers,
                                index=row_headers)
        statement = self.buildStatement(data)
        self.debug("seaborn: executing statement '%s'" % statement)
        plot = self.execute(statement, globals(), locals())
        self.debug("plot matrix finished")
        return plot


class HeatmapPlot(SeabornMatrixPlot):
    """Render a matrix as a heatmap using seaborn.
    """

    def buildStatement(self, data):
        return("plot = seaborn.heatmap(data)")


class ClustermapPlot(SeabornMatrixPlot):
    """Render a matrix as a heatmap using seaborn.
    """

    options = (('row-regex', directives.unchanged),
               ('col-regex', directives.unchanged),
               ) + SeabornMatrixPlot.options

    row_regex = None
    col_regex = None

    def __init__(self, *args, **kwargs):
        SeabornMatrixPlot.__init__(self, *args, **kwargs)
        self.row_regex = kwargs.get("row-regex", None)
        self.col_regex = kwargs.get("col-regex", None)

    def buildStatement(self, data):

        if data.isnull().any().any():
            raise ValueError("dataframe contains NaN")

        extra_options = []
        if self.row_regex:
            row_names = data.index.str.extract(self.row_regex)
            factors = list(set(row_names))
            colors = matplotlib.cm.rainbow(numpy.linspace(0, 1, len(factors)))
            factor2color = dict([(y, x) for x, y in enumerate(factors)])
            row_colors = [colors[factor2color[x]] for x in row_names]

            extra_options.append(
                "row_colors={}".format(str(row_colors)))

        if extra_options:
            extra_options = "," + ",".join(extra_options)
        else:
            extra_options = ""

        return("plot = seaborn.clustermap(data {})".format(extra_options))



import numpy
import pandas
import re
from collections import OrderedDict
from six import string_types, integer_types

ContainerTypes = (tuple, list, type(numpy.zeros(0)))
DictionaryTypes = (dict, OrderedDict)

# Taken from numpy.ScalarType, but removing the types object and unicode
# None is allowed to represent missing values. numpy.float128 is a recent
# numpy addition.
try:
    FloatTypes = (float,
                  numpy.float32, numpy.float64, numpy.float128)
    IntTypes = integer_types + (
        numpy.int8, numpy.int16,
        numpy.int32, numpy.int64,
        numpy.uint8, numpy.uint16,
        numpy.uint32, numpy.uint64)

except AttributeError as msg:
    FloatTypes = (float,
                  numpy.float32, numpy.float64)
    IntTypes = integer_types + (
        numpy.int8, numpy.int16,
        numpy.int32, numpy.int64,
        numpy.uint8, numpy.uint16,
        numpy.uint32, numpy.uint64)

NumberTypes = IntTypes + FloatTypes + (type(None),)


def is_dataframe(data):
    '''return True if data is a dataframe.'''
    return type(data) == pandas.DataFrame


def is_dataseries(data):
    '''return True if data is a series.'''
    return type(data) == pandas.Series


def is_array(data):
    '''return True if data is an array.'''
    return type(data) in ContainerTypes


def is_matrix(data):
    '''return True if data is a numpy matrix.

    A matrix is an array with two dimensions.
    '''
    return isinstance(data, numpy.ndarray) and len(data.shape) == 2


def is_dict(data):
    '''return True if data is a dictionary'''
    return type(data) in DictionaryTypes


def is_int(obj):
    return type(obj) in IntTypes


def is_float(obj):
    return type(obj) in FloatTypes


def is_string(obj):
    # Python 3
    # return isinstance(obj, str)
    return isinstance(obj, string_types)


def is_numeric(obj):
    attrs = ['__add__', '__sub__', '__mul__', '__div__', '__pow__']
    return all(hasattr(obj, attr) for attr in attrs)


def as_list(param):
    '''return a param as a list'''
    if type(param) not in (list, tuple):
        p = param.strip()
        if p:
            return [x.strip() for x in p.split(",")]
        else:
            return []
    else:
        return param


def quote_rst(text):
    '''quote text for restructured text.'''
    return re.sub(r"([*])", r"\\\1", str(text))


def quote_filename(text):
    '''quote filename for use as link in restructured text (remove spaces,
    quotes, slashes, etc).

    latex does not permit a "." for image files.

    Note that the quoting removes slashes and backslashes and thus removes
    any path information.

    Replace all with "_"

    '''
    return re.sub(r"""[ '"()\[\]./]""", r"_", str(text))


def get_encoding():
    # TODO: hard-coded for now, parameterize later.
    return "utf-8"


def force_decode(s, encoding="utf-8", errors="replace"):
    if s is None:
        return None
    return s.decode(encoding, errors=errors)


def force_encode(s, encoding="utf-8", errors="replace"):
    if s is None:
        return None
    return s.encode(encoding, errors=errors)


def force_dataframe_encode(df, encoding="utf-8", errors="replace"):
    df = df.reset_index()
    columns = [x for x, y in zip(df.columns, df.dtypes) if y == object]
    for col in columns:
        try:
            df[col] = df[col].str.decode("unicode-escape").str.encode(
                encoding, errors=errors)
        except AttributeError:
            # non-string columns
            pass
    return df



import os
import shelve
import pickle
import six

from CGATReport.Component import Component
from CGATReport import Utils
from CGATReport.Types import force_encode


# Python 3 - bsddb.db not available
# import bsddb.db


def tracker2key(tracker):
    '''derive cache filename from a tracker.'''

    modulename = tracker.__module__
    try:
        # works for functions (def)
        name = tracker.__name__
    except AttributeError:
        # works for functors (class)
        name = tracker.__class__.__name__

    return Utils.quote_filename(".".join((modulename, name)))


class Cache(Component):

    '''persistent storage for tracker results.'''

    def __init__(self, cache_name, mode="a"):

        Component.__init__(self)

        self.cache_filename = None
        self._cache = None
        self.cache_name = cache_name
        if "report_cachedir" in Utils.PARAMS:
            self.cache_dir = Utils.PARAMS["report_cachedir"]
        else:
            self.cache_dir = None

        if self.cache_dir:

            try:
                os.mkdir(self.cache_dir)
            except OSError as msg:
                pass

            if not os.path.exists(self.cache_dir):
                raise OSError("could not create directory %s: %s" %
                              (self.cache_dir, msg))

            self.cache_filename = os.path.join(self.cache_dir,
                                               cache_name)

            if mode == "r":
                if not os.path.exists(self.cache_filename):
                    raise ValueError("cache %s does not exist at %s" %
                                     (self.cache_name,
                                      self.cache_filename))

            # on Windows XP, the shelve does not work, work without cache
            try:
                self._cache = shelve.open(
                    self.cache_filename, "c", writeback=False)
                self.debug("disp%s: using cache %s" %
                           (id(self), self.cache_filename))
                self.debug("disp%s: keys in cache: %s" %
                           (id(self,), str(list(self._cache.keys()))))
            # except bsddb.db.DBFileExistsError as msg:
            except OSError as msg:
                self.warn(
                    "disp%s: could not open cache %s - continuing without. Error = %s" %
                    (id(self), self.cache_filename, msg))
                self.cache_filename = None
                self._cache = None
        else:
            self.debug("disp%s: not using cache" % (id(self),))

    def __del__(self):

        if self._cache is None:
            return
        self.debug("closing cache %s" % self.cache_filename)
        self.debug("keys in cache %s" % (str(self._cache.keys())))
        self._cache.close()
        self._cache = None

    def keys(self):
        '''return keys in cache.'''
        if self._cache is not None:
            return sorted(self._cache.keys())
        else:
            return []

    def __getitem__(self, key):
        '''return data in cache.
        '''

        if self._cache is None:
            raise KeyError("no cache - key `%s` does not exist" % str(key))

        if six.PY2:
            key = force_encode(key)

        try:
            if key in self._cache:
                result = self._cache[key]
                if result is not None:
                    self.debug(
                        "retrieved data for key '%s' from cache" % (key))
                else:
                    self.warn(
                        "retrieved None data for key '%s' from cache" % (key))
            else:
                self.debug("key '%s' not found in cache" % key)
                raise KeyError("cache does not contain %s" % str(key))

        # except (bsddb.db.DBPageNotFoundError, bsddb.db.DBAccessError,
        # pickle.UnpicklingError, ValueError, EOFError) as msg:
        except (pickle.UnpicklingError, ValueError, EOFError) as msg:
            self.warn("could not get key '%s' or value for key in '%s': msg=%s" %
                      (key,
                       self.cache_filename,
                       msg))
            raise KeyError("cache could not retrieve %s" % str(key))

        return result

    def __setitem__(self, key, data):
        '''save data in cache.
        '''
        if self._cache is None:
            return

        if six.PY2:
            key = force_encode(key)
        else:
            if isinstance(key, bytes):
                raise ValueError("key '{}' is a byte string".format(key))

        # do not store data frames in cache until
        # best method is clear
        try:
            self._cache[key] = data
            self.debug("saved data for key '%s' in cache" % key)
        # except (bsddb.db.DBPageNotFoundError,bsddb.db.DBAccessError) as
        # msg:
        except (OSError) as msg:
            self.warn("could not save key '%s' from '%s': msg=%s" %
                      (key,
                       self.cache_filename,
                       msg))
        # The following sync call is absolutely necessary when
        # using the multiprocessing library (python
        # 2.6.1). Otherwise the cache is emptied somewhere before
        # the final call to close(). Even necessary, if writeback
        # = False
        self._cache.sync()



import re
import os
import sys
import traceback
import math
import glob
import pkgutil
import pandas

from logging import warning
from functools import reduce
import multiprocessing

# Python 2/3 Compatibility
try:
    import ConfigParser as configparser
except:
    import configparser


import CGATReport
from CGATReport.ResultBlock import ResultBlocks, ResultBlock
import CGATReport.Component as Component
import CGATReport.Config
from CGATReport.Types import quote_filename, quote_rst, is_string, as_list

# set with keywords that will not be pruned
# This is important for the User Tracker
TrackerKeywords = set(("text", "rst", "xls",))

# Options for rst image directive that will get passed through
# unchanged.
ImageOptions = set(("alt", "height", "width", "scale", "class"))


# default values
PARAMS = {
    "report_show_errors": True,
    "report_show_warnings": True,
    "report_sql_backend": "sqlite:///./csvdb",
    "report_cachedir": "_cache",
    "report_urls": "data,code,rst",
    "report_images": "hires,hires.png,200,eps,eps,50",
}


# read placeholders from config file in current directory
# It would be nice to read default values, but the location
# of the documentation source files are not known to this module.
class memoized(object):

    """Decorator that caches a function's return value each time it is called.
    If called later with the same arguments, the cached value is returned, and
    not re-evaluated.
    """

    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        try:
            return self.cache[args + list(kwargs)]
        except KeyError:
            self.cache[
                args + list(kwargs)] = value = self.func(*args, **kwargs)
            return value
        except TypeError:
            # uncachable -- for instance, passing a list as an argument.
            # Better to not cache than to blow up entirely.
            return self.func(*args, **kwargs)

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__


def getDataFrameLevels(dataframe,
                       test_for_trivial=False):
    '''return numbers of levels in the index
    of the dataframe.

    A non-hierarchical index has a level of 1.

    If test_for_trivial is set to true and the index
    is not hierarchical, return 0 if the index is
    simply the row numbers.

    '''
    index = dataframe.index
    try:
        # hierarchical index
        nlevels = len(index.levels)
    except AttributeError:
        nlevels = 1
        if test_for_trivial and isinstance(index, pandas.Int64Index) \
           and index.name is None:
            nlevels = 0
    return nlevels


def getGroupLevels(dataframe,
                   max_level=None,
                   modify_levels=None):
    '''return expression for pandas groupby statement.

    If dataframe has a multiindex, return a tuple
    of levels.

    If *modify_levels* is set to *n*, the tuple of group_levels
    will be reduced:
    If n is > 0, the first n levels are dropped. If n is negative,
    the last n levels are dropped.

    If dataframe is not a multiindex, return a level
    of 0.
    '''
    nlevels = getDataFrameLevels(dataframe)
    if nlevels == 1:
        return 0
    else:
        if max_level is not None:
            return tuple(range(min(max_level, nlevels)))
        elif modify_levels is not None:
            if modify_levels > 0:
                l = list(range(modify_levels, nlevels))
            else:
                l = list(range(0, nlevels + modify_levels))
            return tuple(l)
        else:
            return tuple(range(nlevels))


def pruneDataFrameIndex(dataframe,
                        expected_levels=None,
                        original=None):
    '''prune levels in index of a dataframe.

    If *original* is given, prune to the same
    number of levels as in *original*.

    This function is used after a pandas.concat
    operation. If the concatenated dataframes
    contain indices, they will be in the final
    result.

    This method modifies the dataframe in-place.
    '''
    if original is not None:
        expected_levels = original.index.nlevels

    if expected_levels is not None:
        nlevels = dataframe.index.nlevels
        if nlevels > expected_levels:
            dataframe.reset_index(
                level=list(range(expected_levels,
                                 nlevels)),
                drop=True,
                inplace=True)


def toMultipleSeries(dataframe):
    '''split a dataframe with a hierarchical
    index into separated data series.'''

    level = getGroupLevels(dataframe)
    dataseries = []
    for key, work in dataframe.groupby(
            level=level):
        for column in work.columns:
            dataseries.append((key + (column,), work[column]))
    return dataseries


def convertValue(value, list_detection=False):
    '''convert a value to int, float or str.'''
    rx_int = re.compile("^\s*[+-]*[0-9]+\s*$")
    rx_float = re.compile("^\s*[+-]*[0-9.]+[.+\-eE][+-]*[0-9.]*\s*$")

    if value is None:
        return value

    if list_detection and "," in value:
        values = []
        for value in value.split(","):
            if rx_int.match(value):
                values.append(int(value))
            elif rx_float.match(value):
                values.append(float(value))
            else:
                values.append(value)
        return values
    else:
        if rx_int.match(value):
            return int(value)
        elif rx_float.match(value):
            return float(value)
        return value


def configToDictionary(config):
    p = {}
    for section in config.sections():
        for key, value in config.items(section):
            v = convertValue(value)
            p["%s_%s" % (section, key)] = v
            if section == "general":
                p["%s" % (key)] = v

    for key, value in list(config.defaults().items()):
        p["%s" % (key)] = convertValue(value)

    return p


def update_parameters(filenames):
    '''read one or more config files and update parameter
    dictionary.

    Sections and keys are combined with an underscore. If a key
    without section does not exist, it will be added plain.

    For example::

       [general]
       input=input1.file

       [special]
       input=input2.file

    will be entered as { 'general_input': "input1.file",
    'input: "input1.file", 'special_input': "input2.file" }

    This function also updates the module-wide parameter map.

    The section [DEFAULT] is equivalent to [general].

    '''

    global CONFIG

    CONFIG = configparser.ConfigParser()
    CONFIG.read([x for x in filenames if x])

    p = configToDictionary(CONFIG)
    PARAMS.update(p)

    return PARAMS

# DEFAULT: read parameters from config files
# in the current directory sorted by name
# If 'inifile' is in conf.py, read it first.


def get_parameters():
    return update_parameters(
        filenames=[getattr(CGATReport.Config, 'inifile', None)] +
        sorted(glob.glob("*.ini")))


def getImageFormats(display_options=None):
    '''return list of image formats to render (in addition to the default
    format).
    '''

    def _toFormat(data):
        if len(data) == 1:
            return (data[0], data[0], 100)
        elif len(data) == 2:
            return (data[0], data[1], 100)
        elif len(data) == 3:
            return (data[0], data[1], int(data[2]))
        else:
            raise ValueError(
                ":format: option expects one to three params, "
                "got {}".format(data))

    if "display" in display_options:
        all_data = [x.strip() for x in display_options["display"].split(";")]
        if len(all_data) > 1:
            warning(":display: only expects one format, additional ignored at %s" %
                    display_options["display"])
        data = as_list(all_data[0])
        default_format = _toFormat(data)
    elif "report_default_format" in PARAMS:
        default_format = _toFormat(PARAMS["report_default_format"].split(","))
    else:
        default_format = CGATReport.Config.HTML_IMAGE_FORMAT

    # get default extra formats from the config file
    additional_formats = []
    if "report_images" in PARAMS:
        data = as_list(PARAMS["report_images"])
        if len(data) % 3 != 0:
            raise ValueError(
                "need multiple of 3 number of arguments to report_images "
                "option, got {}".format(data))
        for x in range(0, len(data), 3):
            additional_formats.append((data[x], data[x + 1], int(data[x + 2])))
    else:
        additional_formats.extend(CGATReport.Config.ADDITIONAL_FORMATS)

    # add formats specified in the document
    if "extra-formats" in display_options:
        all_data = [x.strip() for x in
                    display_options["extra-formats"].split(";")]
        for data in all_data:
            data = as_list(all_data[0])
            additional_formats.append(_toFormat(data))

    if CGATReport.Config.LATEX_IMAGE_FORMAT:
        additional_formats.append(CGATReport.Config.LATEX_IMAGE_FORMAT)

    return default_format, additional_formats


def get_default_display_options():
    """return dictionary with default display options from the config file.
    """
    display_options = {}
    if "report_default_width" in PARAMS:
        display_options["width"] = PARAMS["report_default_width"]
    return display_options


def getImageOptions(display_options=None, indent=0):
    '''return string with display_options for ``:image:`` directive.'''
    indent = ' ' * indent
    if display_options:
        return "\n".join(
            ['%s:%s: %s' % (indent, key, val)
             for key, val in list(display_options.items())
             if key in ImageOptions])
    else:
        return ''


def indent(text, indent):
    '''return *text(indented by *indent*.'''
    return "\n".join([" " * indent + x for x in text.split("\n")])


def buildWarning(name, message):
    '''build a cgatreport warning message with name *name*
    and message *message*.

    Note that *name* should not contain any spaces.

    '''
    if PARAMS["report_show_warnings"]:
        WARNING_TEMPLATE = '''

.. warning:: %(name)s

   * %(message)s

'''
        name = re.sub("\s", "_", name)
        return ResultBlock(WARNING_TEMPLATE % locals(),
                           title="")
    else:
        return None


def buildException(stage):
    '''build an exception text element.

    It uses the last exception.
    '''

    if PARAMS["report_show_errors"]:
        EXCEPTION_TEMPLATE = '''
.. error:: %(exception_name)s

   * stage: %(stage)s
   * exception: %(exception_name)s
   * message: %(exception_value)s
   * traceback:

%(exception_stack)s
'''

        exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
        lines = traceback.format_tb(exceptionTraceback)
        # remove all the ones relate to Dispatcher.py
        # xlines = filter(lambda x: not re.search("Dispatcher.py", x), lines)
        xlines = lines
        # if nothing left, use the full traceback
        if len(xlines) > 0:
            lines = xlines
        # add prefix of 6 spaces
        prefix = "\n" + " " * 6
        exception_stack = quote_rst(prefix +
                                    prefix.join("".join(lines).split("\n")))
        if exceptionType.__module__ == "exceptions":
            exception_name = quote_rst(exceptionType.__name__)
        else:
            exception_name = quote_rst(exceptionType.__module__ + '.' +
                                       exceptionType.__name__)

        exception_value = quote_rst(str(exceptionValue))

        return ResultBlock(EXCEPTION_TEMPLATE % locals(),
                           title="")
    else:
        return None


def collectExceptionAsString(msg):
    '''return exception as a string.'''
    exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
    lines = traceback.format_tb(exceptionTraceback)
    return 'exception: %s\nmsg=%s\n%s' % (str(exceptionType),
                                          msg,
                                          "".join(lines))


def my_get_data(package, resource):
    """Get a resource from a package.

    This is a wrapper round the PEP 302 loader get_data API. The package
    argument should be the name of a package, in standard module format
    (foo.bar). The resource argument should be in the form of a relative
    filename, using '/' as the path separator. The parent directory name '..'
    is not allowed, and nor is a rooted name (starting with a '/').

    The function returns a binary string, which is the contents of the
    specified resource.

    For packages located in the filesystem, which have already been imported,
    this is the rough equivalent of

        d = os.path.dirname(sys.modules[package].__file__)
        data = open(os.path.join(d, resource), 'rb').read()

    If the package cannot be located or loaded, or it uses a PEP 302 loader
    which does not support get_data(), then None is returned.
    """

    loader = pkgutil.get_loader(package)
    if loader is None or not hasattr(loader, 'get_data'):
        return None
    mod = sys.modules.get(package) or loader.load_module(package)
    if mod is None or not hasattr(mod, '__file__'):
        return None

    # Modify the resource name to be compatible with the loader.get_data
    # signature - an os.path format "filename" starting with the dirname of
    # the package's __file__
    parts = resource.split('/')
    parts.insert(0, os.path.dirname(mod.__file__))
    resource_name = os.path.join(*parts)
    return loader.get_data(resource_name)

# get_data only available in python >2.6
try:
    from pkgutil import get_data
except ImportError:
    get_data = my_get_data


def getTemplatesDir():
    '''returns the location of the templates.'''
    return os.path.join(os.path.dirname(__file__), "templates")


def normalize_cell(s, length):
    return s + ((length - len(s)) * ' ')


def table_div(num_cols, col_width, header_flag):
    if header_flag == 1:
        return num_cols * ('+' + (col_width) * '=') + '+\n'
    else:
        return num_cols * ('+' + (col_width) * '-') + '+\n'


def table2rst(table):
    """convert a table (list of lists) to rst table.
    """
    cell_width = 2 + max(
        reduce(lambda x, y: x + y,
               [[max(list(map(len, str(item).split('\n')))) for item in row]
                for row in table], []))
    num_cols = len(table[0])
    rst = table_div(num_cols, cell_width, 0)
    header_flag = 1
    for row in table:
        split_row = [str(cell).split('\n') for cell in row]
        lines_remaining = 1

        while lines_remaining > 0:
            normalized_cells = []
            lines_remaining = 0
            for cell in split_row:
                lines_remaining += len(cell)

                if len(cell) > 0:
                    normalized_cell = normalize_cell(
                        str(cell.pop(0)), cell_width - 1)
                else:
                    normalized_cell = normalize_cell(
                        '', cell_width - 1)

                normalized_cells.append(normalized_cell)

            rst = rst + '| ' + '| '.join(normalized_cells) + '|\n'

        rst = rst + table_div(num_cols, cell_width, header_flag)
        header_flag = 0

    return rst


def layoutBlocks(blocks, layout="column"):
    """layout blocks of rst text.

    layout can be one of "column", "row", or "grid".

    The layout uses an rst table to arrange elements.
    """

    logger = Component.get_logger()

    lines = []
    if len(blocks) == 0:
        return lines

    # flatten blocks
    bb = ResultBlocks()
    for b in blocks:
        if b.title:
            b.updateTitle(b.title, "prefix")
        try:
            bb.extend(b)
        except TypeError:
            bb.append(b)

    blocks = bb

    # check if postambles are identical across all blocks
    postambles = set([b.postamble for b in blocks])

    if len(postambles) == 1:
        blocks.clearPostamble()
        postamble = postambles.pop()
    else:
        postamble = None

    if layout == "column":
        for block in blocks:
            if block.title:
                lines.extend(block.title.split("\n"))
                lines.append("")
            else:
                logger.warning("report_directive.layoutBlocks: missing title")

            lines.extend(block.text.split("\n"))
            lines.extend(block.postamble.split("\n"))

        lines.append("")

        if postamble:
            lines.extend(postamble.split("\n"))
            lines.append("")
        return lines

    elif layout in ("row", "grid"):
        if layout == "row":
            ncols = len(blocks)
        elif layout == "grid":
            ncols = int(math.ceil(math.sqrt(len(blocks))))

    elif layout.startswith("column"):
        ncols = min(len(blocks), int(layout.split("-")[1]))
        # TODO: think about appropriate fix for empty data
        if ncols == 0:
            ncols = 1
            return lines
    else:
        raise ValueError("unknown layout %s " % layout)

    if ncols == 0:
        logger.warn("no columns")
        return lines

    # compute column widths
    widths = [x.getWidth() for x in blocks]
    text_heights = [x.getTextHeight() for x in blocks]
    title_heights = [x.getTitleHeight() for x in blocks]

    columnwidths = []
    for x in range(ncols):
        columnwidths.append(max([widths[y] for y in
                                 range(x, len(blocks), ncols)]))

    separator = "+%s+" % "+".join(["-" * x for x in columnwidths])

    # add empty blocks
    if len(blocks) % ncols:
        blocks.extend([ResultBlock("", "")] * (ncols - len(blocks) % ncols))

    for nblock in range(0, len(blocks), ncols):

        # add text
        lines.append(separator)
        max_height = max(text_heights[nblock:nblock + ncols])
        new_blocks = ResultBlocks()

        for xx in range(nblock, min(nblock + ncols, len(blocks))):
            txt, col = blocks[xx].text.split("\n"), xx % ncols
            txt = blocks[xx].text.split("\n") + \
                blocks[xx].postamble.split("\n")
            col = xx % ncols

            max_width = columnwidths[col]

            # add missig lines
            txt.extend([""] * (max_height - len(txt)))
            # extend lines
            txt = [x + " " * (max_width - len(x)) for x in txt]

            new_blocks.append(txt)

        for l in zip(*new_blocks):
            lines.append("|%s|" % "|".join(l))

        # add subtitles
        max_height = max(title_heights[nblock:nblock + ncols])

        if max_height > 0:

            new_blocks = ResultBlocks()
            lines.append(separator)

            for xx in range(nblock, min(nblock + ncols, len(blocks))):

                txt, col = blocks[xx].title.split("\n"), xx % ncols

                max_width = columnwidths[col]
                # add missing lines
                txt.extend([""] * (max_height - len(txt)))
                # extend lines
                txt = [x + " " * (max_width - len(x)) for x in txt]

                new_blocks.append(txt)

            for l in zip(*new_blocks):
                lines.append("|%s|" % "|".join(l))

    lines.append(separator)

    if postamble:
        lines.append(postamble)

    lines.append("")

    return lines


def get_params():
    return PARAMS


def getOutputDirectory():
    return os.path.join('_static', 'report_directive')


def build_paths(reference):
    '''return paths and filenames for a tracker.

    Reference is usually a Tracker such as "Tracker.TrackerImages".
    '''
    basedir, name = os.path.split(reference)
    # quote name to filename
    filename = quote_filename(name)
    basename, ext = os.path.splitext(filename)
    # note: outdir had basedir at the end?
    outdir = getOutputDirectory()
    codename = os.path.join(basedir, filename) + ".code"
    notebookname = os.path.join(basedir, filename) + ".notebook"

    return basedir, filename, basename, ext, outdir, codename, notebookname

NOTEBOOK_TEMPLATE = """
<!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>
<script src="%(staticdir)s/jquery.min.js"></script>
<script src="%(staticdir)s/notebook.js"></script>
<script>
$(document).ready(function() {
$("#rendered").click(function(){
    create_notebook('%(rendered_options)s');})
});
$(document).ready(function() {
$("#data").click(function(){
    create_notebook('%(data_options)s');})
});
</script>

</head>
<body>

<h1>Notebook integration</h1>

<p>
Click on the buttons below to import the figure
or the data into your notebook.
</p>
<button type="button" id="data">Data</button>
<button type="button" id="rendered">Figure</button>

<h2>Copy and Paste</h2>
Copy and paste the text below for importing the data
in your notebook without rendering.

<pre>
%%matplotlib inline
import CGATReport.test
result = CGATReport.test.main(%(data_options)s)
</pre>

Copy and paste below for importing the rendered
display item in your notebook.
<pre>
%%matplotlib inline
import CGATReport.test
result = CGATReport.test.main(%(rendered_options)s)
</pre>

</body>
</html>
"""

NOTEBOOK_TEXT_TEMPLATE = """
%%matplotlib inline
import CGATReport.test
result = CGATReport.test.main(%(options)s)
"""


def writeNoteBookEntry(outfile, tracker, renderer, transformers,
                       tracker_path, options):
    '''output text for pasting into an ipython notebook into *outfile*
    '''

    logger = Component.get_logger()

    cmd_options = [
        'do_print = False',
        'tracker="%s"' % tracker,
        'trackerdir="%s"' % os.path.dirname(tracker_path),
        'workdir="%s"' % os.getcwd()]

    for key, val in options:
        if val is None:
            cmd_options.append("%s" % key)
        else:
            if is_string(val):
                cmd_options.append('%s="%s"' % (key, val))
            else:
                cmd_options.append("%s=%s" % (key, str(val)))

    if transformers:
        cmd_options.append('transformers=["%s"]' % '","'.join(transformers))

    # For the javascript to work, the options must not contain any
    # single quotes.
    rendered_options = ",".join(cmd_options + ['renderer="%s"' % renderer])
    data_options = ",".join(cmd_options + ['renderer="none"'])
    if "'" in rendered_options:
        logger.warning("notebook options contain ': %s" % (rendered_options))

    # no module name in tracker
    # javascript files life in the "_static" directory, while
    # the notebook entry will end up in "_static/report_directive"
    params = {"rendered_options": rendered_options,
              "data_options": data_options,
              "curdir": os.getcwd(),
              "staticdir": ".."}

    outfile.write(NOTEBOOK_TEMPLATE % params)


def buildRstWithImage(outname,
                      outdir, rstdir, builddir, srcdir,
                      additional_formats,
                      tracker_id,
                      links,
                      display_options,
                      default_format=None,
                      is_html=False,
                      text=None):
    '''output rst text for inserting an image.'''
    rst_output = ""

    # path to build directory from rst directory
    rst2builddir = os.path.join(os.path.relpath(builddir,
                                                start=rstdir), outdir)

    # path to src directory from rst directory
    rst2srcdir = os.path.join(os.path.relpath(srcdir,
                                              start=rstdir), outdir)

    # for image directive - image path is relative from rst file to
    # external build dir
    imagepath = re.sub("\\\\", "/",
                       os.path.join(rst2builddir,
                                    outname))

    # for links - path is from rst file to internal root dir
    relative_imagepath = re.sub("\\\\", "/",
                                os.path.join(
                                    rst2srcdir,
                                    outname))

    requested_urls = as_list(PARAMS["report_urls"])

    image_options = getImageOptions(display_options, indent=6)

    if CGATReport.Config.HTML_IMAGE_FORMAT:

        if default_format:
            id, format, dpi = default_format
        else:
            id, format, dpi = CGATReport.Config.HTML_IMAGE_FORMAT

        #######################################################
        # Define templates
        if text is not None:

            template = '''
.. only:: html

   .. raw:: html

      <div><p>
         %(outname)s
      <p><div>

   %(url)s
'''
            # put text in outname
            outname = indent(text, 6)

        elif is_html:
            # put in html
            template = '''
.. only:: html

   .. raw:: html
      :file: %(outname)s

   %(url)s
'''
            # use absolute path for html file
            outname = os.path.abspath(os.path.join(outdir, outname)) + ".html"

        else:
            # put in image directive
            template = '''
.. only:: html

   .. image:: %(linked_image)s
%(image_options)s

   %(url)s

.. only:: pdf

   .. image:: %(linked_image)s
%(image_options)s
'''
            linked_image = imagepath + ".%s" % format
        #######################################################

        extra_images = []
        for id, format, dpi in additional_formats:
            extra_images.append(
                ":download:`%(id)s <%(imagepath)s.%(format)s>`" %
                locals())
        if extra_images:
            extra_images = " " + " ".join(extra_images)
        else:
            extra_images = ""

        # construct additional urls
        urls = []
        if "code" in requested_urls:
            urls.append(":download:`code <%(code_url)s>`" % links)
        if "notebook" in requested_urls:
            urls.append(":download:`nb <%(notebook_url)s>`" % links)
        if "data" in requested_urls:
            urls.append(":download:`data </data/%(tracker_id)s>`" % locals())
        if "rst" in requested_urls and links["rst_url"] is not None:
            urls.append(":download:`rst <%(rst_url)s>`" % links)
        if extra_images:
            urls.append(extra_images)

        if urls and "no-links" not in display_options:
            url = "[{}]".format(" ".join(urls))
        else:
            url = ""

        rst_output += template % locals()

    # treat latex separately
    if CGATReport.Config.LATEX_IMAGE_FORMAT:
        id, format, dpi = CGATReport.Config.LATEX_IMAGE_FORMAT
        template = '''
.. only:: latex

   .. image:: %(linked_image)s
%(image_options)s
'''
        linked_image = imagepath + ".%s" % format
        rst_output += template % locals()

    return rst_output



import multiprocessing
import logging
import sys
import re
import os
import io
import threading
import time
import collections

# Python 2/3 Compatibility
try:
    import queue
except ImportError:
    import Queue as queue

from logging import Logger


class MultiProcessingLogHandler(logging.Handler):

    """taken from http://stackoverflow.com/questions/641420/how-should-i-log-while-using-multiprocessing-in-python

    added counting of log messages.
    """

    def __init__(self, handler, queue, child=False):
        logging.Handler.__init__(self)

        self._handler = handler
        self.queue = queue

        self.counts = collections.defaultdict(int)

        # we only want one of the loggers to be pulling from the queue.
        # If there is a way to do this without needing to be passed this
        # information, that would be great!
        if child == False:
            self.shutdown = False
            self.polltime = 1
            t = threading.Thread(target=self.receive)
            t.daemon = True
            t.start()

    def setFormatter(self, fmt):
        logging.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def receive(self):
        while (self.shutdown == False) or (self.queue.empty() == False):
            # so we block for a short period of time so that we can
            # check for the shutdown cases.
            try:
                record = self.queue.get(True, self.polltime)
                self._handler.emit(record)
                self.counts[record.levelname] += 1
            except queue.Empty as e:
                pass
            except EOFError as e:
                break

    def send(self, s):
        # send just puts it in the queue for the server to retrieve
        self.queue.put(s)

    def _format_record(self, record):
        ei = record.exc_info
        if ei:
            # just to get traceback text into record.exc_text
            dummy = self.format(record)
            record.exc_info = None  # to avoid Unpickleable error

        return record

    def emit(self, record):
        try:
            s = self._format_record(record)
            self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def close(self):
        # give some time for messages to enter the queue.
        time.sleep(self.polltime + 1)
        self.shutdown = True
        # give some time for the server to time out and see the shutdown
        time.sleep(self.polltime + 1)

    def __del__(self):
        # hopefully this aids in orderly shutdown when things are going poorly.
        self.close()

    def getCounts(self):
        return self.counts


def f(x):
    # just a logging command...
    logging.critical('function number: ' + str(x))
    # to make some calls take longer than others, so the output is "jumbled"
    # as real MP programs are.
    time.sleep(x % 3)


def initPool(queue, level):
    """
    This causes the logging module to be initialized with the necessary info
    in pool threads to work correctly.
    """
    logging.getLogger('').addHandler(
        MultiProcessingLogHandler(logging.StreamHandler(), queue, child=True))
    logging.getLogger('').setLevel(level)

if __name__ == '__main__':

    use_stream = False

    if use_stream:
        stream = io.StringIO()
        logQueue = multiprocessing.Queue(100)
        handler = MultiProcessingLogHandler(
            logging.StreamHandler(stream), logQueue)
        logging.getLogger('').addHandler(handler)
        logging.getLogger('').setLevel(logging.DEBUG)

        logging.debug('starting main')

        # when bulding the pool on a Windows machine we also have to init the
        # logger in all the instances with the queue and the level of logging.
        pool = multiprocessing.Pool(processes=10, initializer=initPool, initargs=[
                                    logQueue, logging.getLogger('').getEffectiveLevel()])  # start worker processes
        pool.map(f, list(range(0, 50)))
        pool.close()

        logging.debug('done')
        logging.shutdown()
        print("stream output is:")
        print((stream.getvalue()))
    else:
        logQueue = multiprocessing.Queue(100)
        handler = MultiProcessingLogHandler(
            logging.FileHandler("test.log", "w"), logQueue)
        # logging.setLoggerClass(CountedLogger)
        #logging.root = CountedLogger("")
        # print dir(logging.getLogger(''))
        # print logging.getLogger("").getCounts()
        logging.getLogger('').addHandler(handler)
        logging.getLogger('').setLevel(logging.DEBUG)

        logging.debug('starting main')

        # when bulding the pool on a Windows machine we also have to init the
        # logger in all the instances with the queue and the level of logging.
        pool = multiprocessing.Pool(processes=10, initializer=initPool, initargs=[
                                    logQueue, logging.getLogger('').getEffectiveLevel()])  # start worker processes
        pool.map(f, list(range(0, 10)))
        pool.close()

        logging.debug('done')
        logging.shutdown()

        print((handler.getCounts()))



#!/bin/env python
import sys
import re
import glob
import os
import optparse

from CGATReport import Utils

USAGE = """%s [OPTIONS]

set up an new cgatreport in the current directory.
""" % sys.argv[0]


def main(argv=None):

    if argv is None:
        argv = sys.argv

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-d", "--dest", dest="destination", type="string",
                      help="destination directory.")

    parser.set_defaults(
        destination=".",
    )

    (options, args) = parser.parse_args()

    dest = options.destination

    # create directories
    for d in ("",
              "_templates",
              "cgat",
              "cgat/static",
              "trackers",
              "templates"):
        dd = os.path.join(dest, d)
        if not os.path.exists(dd):
            os.makedirs(dd)

    # copy files
    def copy(src, dst):
        fn = os.path.join(dest, dst, os.path.basename(src))
        if os.path.exists(fn):
            raise OSError("file %s already exists - not overwriting." % fn)

        outfile = open(fn, "w")
        x = Utils.get_data("CGATReport", "templates/%s" % src)
        if len(x) == 0:
            raise ValueError('file %s is empty' % src)
        outfile.write(x)
        outfile.close()

    for f in ("Makefile",
              "server.py",
              "cgatreport.ini",
              "conf.py",
              "contents.rst"):
        copy(f, "")

    # for f in ("indexsidebar.html",
    #           "layout.html",
    #           "search.html"):
    #     copy(f, "_templates")

    # for f in ("data_table.html",):
    #     copy(f, "templates")

    for f in ("theme.conf", "layout.html"):
        copy(f, "cgat")

    for f in ("cgat.css_t",
              "js/sorttable.js",
              "js/jquery.dataTables.min.js",
              "js/notebook.js",
              "js/jssor.js",
              "js/jssor.slider.js"):
        copy(f, os.path.join("cgat", "static"))

    for f in ("Trackers.py", ):
        copy(f, "trackers")

    print(("""
Welcome to CGATReport.

All files have been successfully copied to `%(dest)s`. In order to build the
report, go to `%(dest)s`

cd %(dest)s

Optionally edit the configuration in *conf.py* as appropriate:

vi conf.py

As a first attempt, try to build the skeleton book:

make html

If all works, you can start adding text to files in the
directories `analysis` and `pipeline` and pipeline.rst in the
main directory. Add code to collect data to the module 'Trackers.py'
in the  'trackers' directory. If you don't like the default file layout,
it can be easily changed.

""" % locals()))


if __name__ == "__main__":
    sys.exit(main())



import os
from CGATReport.Component import Component

try:
    import bokeh.embed
    import bokeh.plotting
    HAS_BOKEH = True
except ImportError:
    HAS_BOKEH = False


class BokehPlugin(Component):
    '''collect bokeh plot objects and write
    javascript snippets to be included
    raw into html.
    '''

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

    def collect(self,
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links={}):
        '''collect rst output from result blocks.

        '''

        map_figure2text = {}

        if not HAS_BOKEH:
            return map_figure2text

        figid = 10

        for xblocks in blocks:
            for block in xblocks:
                if not hasattr(block, "bokeh"):
                    continue
                figid = block.bokeh._id
                res = bokeh.resources.CDN
                script_path = os.path.join(
                    '_static/report_directive/', "%s.js" % figid)

                js_txt, script_txt = bokeh.embed.autoload_static(
                    block.bokeh, res, script_path)

                # make sure to add the '/' at the end
                # script = block.bokeh.create_html_snippet(
                #     server=False,
                #     embed_base_url='_static/report_directive/',
                #     embed_save_loc=outdir,
                #     static_path='_static/')

                with open(script_path, "w") as outf:
                    outf.write(js_txt)

                text = """.. raw:: html

   <div><p>
         %s
   </p></div>
""" % script_txt

                map_figure2text["#$bkh %s$#" % figid] = text
        return map_figure2text



'''Plotting using the ggplot module.'''

from CGATReport.Plugins.Renderer import Renderer
from CGATReport.Plugins.Plotter import Plotter
from docutils.parsers.rst import directives

import matplotlib.pyplot as plt
import pandas


class PandasPlot(Renderer, Plotter):

    """Use the python ggplot libary for plotting.
    """
    options = (
        ('statement',  directives.unchanged),
        ('kwargs', directives.unchanged)
    ) + Renderer.options + Plotter.options

    group_level = 0

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)
        Plotter.__init__(self, *args, **kwargs)

        # statement is synonym for kwargs. The latter takes precedence
        self.kwargs = kwargs.get('kwargs', kwargs.get('statement'))

    def render(self, dataframe, path):

        # Used to call reset_index() here in order to add
        # the index as a column, but now let caller do this.
        p = None
        s = "p = dataframe.plot(%s)" % self.kwargs

        try:
            exec(s, globals(), locals())
        except Exception as msg:
            raise Exception(
                "pandas.plot() raised error for statement '%s': msg=%s" %
                (s, msg))

        self.mFigure += 1
        if self.title:
            plt.title(self.title)
        plts = [p]
        return self.endPlot(plts, None, path)



import os
import re
from CGATReport.Component import *
from CGATReport import Config


class XLSPlugin(Component):

    capabilities = ['collect']

    def __init__(self, *args, **kwargs):
        Component.__init__(self, *args, **kwargs)

    def collect(self,
                blocks,
                template_name,
                outdir,
                rstdir,
                builddir,
                srcdir,
                content,
                display_options,
                tracker_id,
                links={}):
        '''collect xls output from result blocks.

        xls output is written to a file and a link will be inserted at
        the place holder.
        '''
        map_figure2text = {}
        extension = "xlsx"
        for xblocks in blocks:
            for block in xblocks:
                if not hasattr(block, "xls"):
                    continue

                outname = "%s_%s" % (
                    template_name, re.sub("/", "@", block.title))
                outputpath = os.path.join(outdir, '%s.%s' %
                                          (outname, extension))

                # save to file
                block.xls.save(outputpath)

                # use absolute path
                link = os.path.abspath(outputpath)

                rst_output = "%(link)s" % locals()
                map_figure2text["#$xls %s$#" % block.title] = rst_output

        return map_figure2text



#!/bin/env python

"""
cgatreport-profile
====================

:command:`cgatreport-profile` examines the log file and
computes some summary statistics on rendering times.

   cgatreport-profile

The full list of command line options is listed by suppling:option:`-h/--help`
on the command line. The options are:

**-s/--section** choice
   Only examine performance of certain aspects of cgatreport. Possible choices are
   ``tracker``, ``rst``, and ``renderer``.

**-t/--time** choice
   Report times either as ``milliseconds`` or ``seconds``.

.. note::

   All times are wall clock times.

"""

import sys
import os
import imp
import io
import re
import types
import glob
import optparse
import shutil
import datetime
import logging
import collections

USAGE = """python %s [OPTIONS] target

clean all data according to target.

Targets can contain wild cards.

""" % sys.argv[0]

from CGATReport import Component


class Counter(object):

    '''

    This class stores the calls per source as calls can be made by
    different sources in any order.
    '''

    def __init__(self):
        self._durations = collections.defaultdict(list)
        self._started = collections.defaultdict(int)
        self._calls = collections.defaultdict(int)

    def add(self, started, dt, source=None):
        if self._started[source] != 0 and not started:
            self._durations[source].append(dt - self._started[source])
            self._started[source] = 0
        elif self._started[source] == 0 and started:
            self._calls[source] += 1
            self._started[source] = dt
        else:
            raise ValueError("inconsistent time points, has_started=%s, is_started=%s" % (
                self._started[source], started))

    def reset(self, source=None):
        '''reset last event.'''
        self._started[source] = None
        self._calls[source] = 0

    def getDuration(self):
        if self._durations:
            x = None
            for source, durations in list(self._durations.items()):
                if x == None:
                    x = durations[0]
                for y in durations[1:]:
                    x += y
            return x
        else:
            return datetime.timedelta()

    def getCalls(self):
        return sum(self._calls.values())

    def getRunning(self):
        '''get numbers of tasks unfinished or still running.'''
        return len([x for x, y in list(self._started.items()) if y != 0])

    duration = property(getDuration)
    calls = property(getCalls)
    running = property(getRunning)


def main(argv=None):

    if argv == None:
        argv = sys.argv

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-s", "--section", dest="sections", type="choice", action="append",
                      choices=("tracker", "rst", "renderer"),
                      help="only examine certain sections [default=%default]")

    parser.add_option("-t", "--time", dest="time", type="choice",
                      choices=("seconds", "milliseconds"),
                      help="time to show [default=%default]")

    parser.add_option("-f", "--filter", dest="filter", type="choice",
                      choices=("unfinished", "running", "completed", "all"),
                      help="apply filter to output [default=%default]")

    parser.set_defaults(sections=[],
                        filter="all",
                        time="seconds")

    (options, args) = parser.parse_args()

    rx = re.compile("^[0-9]+")

    if options.sections:
        profile_sections = options.sections
    else:
        profile_sections = ("rst", "tracker", "renderer", "transformer")

    counts = {}
    for section in profile_sections:
        counts[section] = collections.defaultdict(Counter)

    rootpath = os.path.abspath(".")

    if len(args) == 1:
        infile = open(args[0])
    else:
        infile = open(Component.LOGFILE)

    for line in infile:
        if not rx.match(line):
            continue
        data = line[:-1].split()
        if len(data) < 5:
            continue
        date, time, level, source, action = data[:5]

        if action != "profile:":
            continue

        dt = datetime.datetime.strptime(
            " ".join((date, time)), "%Y-%m-%d %H:%M:%S,%f")

        try:
            started, section = data[5:7]
            point = re.sub("object at .*>", "", " ".join(data[7:]))
            if point.startswith("<"):
                point = point[1:]
            point = re.sub(rootpath, "", point)
        except (IndexError, ValueError):
            print((data[5:]))
            print(("malformatted line in logfile: %s" % line))
            continue

        if section.endswith(":"):
            section = section[:-1]

        if source == "report_directive.run:":
            continue
        is_start = started == "started:"

        if source == "build.py:":
            if is_start:
                logging.info("resetting counts at line=%s" % line[:-1])
                counts = {}
                for s in profile_sections:
                    counts[s] = collections.defaultdict(Counter)
            continue

        try:
            counts[section][point].add(is_start, dt, source)
        except ValueError as msg:
            # if there are errors, there is no finish, reset counter
            if is_start:
                counts[section][point].reset(source)
            try:
                counts[section][point].add(is_start, dt, source)
            except ValueError as msg:
                logging.warn("%s: line=%s" % (msg, line))
            except KeyError as msg:
                print(data)
                print(("error in line: (is_start=%s), msg='%s', %s" %
                       (is_start, msg, line)))

    if options.time == "milliseconds":
        f = lambda d: d.seconds + d.microseconds / 1000
    elif options.time == "seconds":
        f = lambda d: d.seconds + d.microseconds / 1000000

    for section in profile_sections:
        sys.stdout.write("\t".join(
            ("section", "object", "ncalls", "duration", "percall", "running")) + "\n")

        running = []
        for objct, c in list(counts[section].items()):

            # apply filters
            if options.filter in ("unfinished", "running") and c.running == 0:
                continue

            d = f(c.duration)
            if c.calls > 0:
                percall = "%6.3f" % (d / float(c.calls))
            else:
                percall = "na"

            sys.stdout.write("\t".join(
                (list(map(str,
                          (section, objct,
                           c.calls,
                           d,
                           percall,
                           c.running,
                           ))))) + "\n")

            running.extend([x for x, y in list(c._started.items()) if y != 0])

        print("running")
        print(("\n".join(map(str, running))))
        sys.stdout.write("\n" * 3)

if __name__ == "__main__":
    sys.exit(main())



from __future__ import unicode_literals

import os
import sys
import re
import json
import six

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from docutils.parsers.rst import directives
import numpy
import pandas
# for output of excel work books
import openpyxl

# import numpy.arange and math.log into namepsace that they
# are available in eval
from numpy import arange
from math import log

from CGATReport.ResultBlock import ResultBlock, ResultBlocks
from CGATReport.DataTree import path2str
from CGATReport import Utils, DataTree
from CGATReport.Types import force_decode, get_encoding, force_dataframe_encode
from CGATReport.Component import Component
from CGATReport import CorrespondenceAnalysis


class Renderer(Component):
    """Base class of renderers that render data into restructured text.

    The subclasses define how to render the data by overloading the
    :meth:`render` method.

    When called, a Renderer and its subclasses will return blocks of
    restructured text. Images are automatically collected from
    matplotlib and inserted at place-holders.

    """
    # plugin fields
    capabilities = ["render"]

    options = (('format', directives.unchanged),
               ('split-at', directives.nonnegative_int),
               ('split-always', directives.unchanged),)

    # required levels in DataTree
    # -1: Renderer can work from any number of levels
    # 0: Renderer needs scalar
    # 1: Renderer requires dataframe with 1 row-label
    # 2: Renderer requires dataframe with 2 row-labels
    # ...
    nlevels = None

    # default number format
    format = "%i"

    # split number of tracks
    split_at = 0

    # tracks always to include if plot is split into
    # several plots.
    split_always = []

    # when splitting, keep first column
    split_keep_first_column = False

    # default group_level for render
    group_level = 0

    # directory where the actual restructured text is located.
    rst_dir = None

    # directory of the root of the source document. rstdir is
    # a subdirectory of this.
    src_dir = None

    # directory in which the document is built.
    build_dir = None

    # dictionary of rst options controlling the display of a
    # directive.
    display_options = {}

    # the build environment
    build_environment = None

    def __init__(self, *args, **kwargs):
        """create an Renderer object using an instance of
        a:class:`Tracker.Tracker`.
        """
        Component.__init__(self, *args, **kwargs)

        self.debug("%s: starting renderer '%s'" %
                   (id(self), str(self)))

        try:
            self.format = kwargs["format"]
        except KeyError:
            pass

        self.split_at = int(kwargs.get("split-at", 0))

        if "split-always" in kwargs:
            self.split_always = kwargs["split-always"].split(',')

    def __call__(self, dataframe, path):
        '''iterate over leaves/branches in data structure.

        This method will call the:meth:`render` method.

        Large dataframes are split into multiple, smaller rendered
        objects if self.split_at is not zero.

        By default, dataframes are split along the hierachical
        index. However, if there is only a single index, but multiple
        columns, the split is performed on the columns instead. This
        is used when splitting coordinate data as a result of the
        histogram transformation.

        '''
        result = ResultBlocks()

        if not self.split_at:
            result.extend(self.render(dataframe, path))
        else:
            # split dataframe at first index
            level = Utils.getGroupLevels(dataframe)
            grouper = dataframe.groupby(level=level)

            # split dataframe column wise if only one index
            # and multiple columns
            if len(grouper) == 1 and len(dataframe.columns) > self.split_at:
                columns = list(dataframe.columns)
                always = []
                if self.split_keep_first_column:
                    always.append(columns[0])
                # columns to always keep
                always.extend([c for c in columns if c in self.split_always])
                columns = [c for c in columns if c not in always]
                for x in range(0, len(columns), self.split_at):
                    # extract a set of columns
                    result.extend(self.render(
                        dataframe.loc[:, always +
                                      columns[x:x + self.split_at]],
                        path))
            # split dataframe along index
            elif len(grouper) >= self.split_at:
                # build groups
                always, remove_always = [], set()

                if self.split_always:
                    for key, work in grouper:
                        for pat in self.split_always:
                            rx = re.compile(pat)
                            if rx.search(path2str(key)):
                                always.append((key, work))
                                remove_always.add(key)

                    grouper = dataframe.groupby(level=level)

                def _group_group(grouper, always, remove_always):
                    group = always[:]
                    for key, work in grouper:

                        if key in remove_always:
                            continue
                        group.append((key, work))

                        if len(group) >= self.split_at:
                            yield group
                            group = always[:]

                    # reconcile index names
                    yield group

                first = True
                for group in _group_group(grouper,
                                          always,
                                          remove_always):
                    # do not plot last dataframe that contains
                    # only the common tracks to plot
                    if not first and len(group) == len(always):
                        continue
                    first = False

                    df = pandas.concat(
                        [x[1] for x in group])

                    # reconcile index names
                    df.index.names = dataframe.index.names
                    result.extend(self.render(df, path))
            else:
                # do not split dataframe
                result.extend(self.render(dataframe, path))

        return result

    def toString(self, value):
        '''returns a number as string

        If not a number, return empty string.'''

        try:
            return self.format % value
        except TypeError:
            return ""
        except ValueError:
            return "nan"

    def set_paths(self, rst_dir, src_dir, build_dir):
        '''set document paths. These are relevant to determine the
        relative position of the object to be rendered in a hierarchy
        of documents.
        '''
        self.rst_dir = rst_dir
        self.src_dir = src_dir
        self.build_dir = build_dir

    def get_paths(self):
        return self.rst_dir, self.src_dir, self.build_dir

    def set_display_options(self, display_options):
        '''set display options given by user.'''
        self.display_options = display_options

    def set_build_environment(self, env):
        '''set the build environment.'''
        self.build_environment = env


class DataFrame(Renderer):
    """
    Renderer outputting the dataframe directly
    as text without any modification.
    """

    options = Renderer.options +\
        (('head', directives.length_or_unitless),
         ('tail', directives.length_or_unitless),
         ('summary', directives.unchanged))

    # output head of dataframe - 0 is no head
    head = 0
    # output tail of dataframe - 0 is no tail
    tail = 0
    # output summary of dataframe
    summary = False

    def __init__(self, *args, **kwargs):

        Renderer.__init__(self, *args, **kwargs)
        self.head = int(kwargs.get('head', 0))
        self.tail = int(kwargs.get('tail', 0))
        self.summary = 'summary' in kwargs

    def __call__(self, dataframe, path):
        result = ResultBlocks()
        texts = []
        if self.head or self.tail:
            if self.head:
                texts.append(str(dataframe.head(self.head)))
            if self.tail:
                texts.append(str(dataframe.tail(self.tail)))
        elif self.summary:
            texts.append(str(dataframe.describe()))
        else:
            texts.append(str(dataframe))

        # add indentation
        texts = ['\n'.join(['   %s' % y for y in x.split('\n')])
                 for x in texts]

        formatted = '''
::

%s

''' % '\n   ...\n'.join(texts)

        result.append(ResultBlock(formatted,
                                  title=path2str(path)))
        return result


class TableBase(Renderer):

    '''base classes for tables and matrices.'''

    options = Renderer.options +\
        (('force', directives.unchanged),
         ('separate', directives.unchanged),
         ('max-rows', directives.length_or_unitless),
         ('large-rows', directives.length_or_unitless),
         ('max-cols', directives.length_or_unitless),
         ('large-html-class', directives.unchanged),
         ('add-rowindex', directives.unchanged),
         ('html-class', directives.unchanged))

    max_rows = 50
    max_cols = 20

    def __init__(self, *args, **kwargs):
        Renderer.__init__(self, *args, **kwargs)

        self.force = "force" in kwargs
        self.separate = "separate" in kwargs
        self.max_rows = kwargs.get("max-rows", 50)
        self.max_cols = kwargs.get("max-cols", 20)
        self.add_rowindex = "add-rowindex" in kwargs
        self.html_class = kwargs.get(
            'html-class',
            Utils.PARAMS.get('report_table_class',
                             None))
        self.large_rows = kwargs.get("max-rows", 20)
        self.large_html_class = kwargs.get(
            'large-html-class',
            Utils.PARAMS.get('report_largetable_class',
                             None))

    def asCSV(self, dataframe, row_headers, col_headers, title):
        '''save the table using CSV.'''

        out = StringIO()
        # use pandas to_csv, which encodes unicode string. Hence
        # go back and forth between unicode->bytes->unicode.

        if six.PY2:
            # In python 2, the csv module does not do unicode.
            # However, the csv module is used by docutils to read the
            # table, so force ASCII here and replace all unicode
            # chars.
            try:
                dataframe.to_csv(out, encoding=get_encoding())
            except UnicodeDecodeError:
                # try really hard
                dataframe = force_dataframe_encode(
                    dataframe.reset_index(),
                    encoding="ascii",
                    errors="replace")
                dataframe.to_csv(out, encoding=get_encoding())

            lines = force_decode(out.getvalue()).split("\n")
        else:
            dataframe.to_csv(out, encoding=get_encoding())
            lines = out.getvalue().split("\n")

        result = []
        result.append(".. csv-table:: %s" % title)

        if len(lines) > self.large_rows:
            if self.large_html_class is not None:
                result.append("   :class: %s" % self.large_html_class)
        else:
            if self.html_class is not None:
                result.append("   :class: %s" % self.html_class)

        if self.add_rowindex:
            raise NotImplementedError('add-rowindex not implemented')
        else:
            result.append('   :header: %s' % lines[0])
            result.append('')
            result.extend(['   %s' % x for x in lines[1:]])

        result.append("")

        return ResultBlock("\n".join(result), title=title)

    def asRST(self, dataframe, row_headers, col_headers, title):
        '''save the table using RST.'''

        out = StringIO()
        dataframe.to_csv(out, encoding=get_encoding())
        data = [x.split(',') for x in out.getvalue().split('\n')]
        # ignore last element - empty
        del data[-1]

        ncols = len(data[0])
        max_widths = [max([len(row[x]) for row in data])
                      for x in range(ncols)]

        separator = "+" + "+".join(["-" * (x + 2) for x in max_widths]) + "+"
        format = "|" + "|".join([" %%%is " % x for x in max_widths]) + "|"

        lines = []
        lines.append(separator)
        for row in data:
            lines.append(format % tuple(map(str, row)))
            lines.append(separator)
        lines.append("")

        return ResultBlock("\n".join(lines), title=title)

    def asFile(self, dataframe, row_headers, col_headers, title):
        '''save the table as HTML file.

        Multiple files of the same Renderer/Tracker combination are
        distinguished by the title.
        '''

        self.debug("%s: saving %i x %i table as file'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))
        lines = []
        lines.append("`%i x %i table <#$html %s$#>`__" %
                     (len(row_headers), len(col_headers),
                      title))

        r = ResultBlock("\n".join(lines) + "\n", title=title)

        out = StringIO()
        dataframe.to_csv(out)
        lines = out.getvalue().split("\n")

        # create an html table
        data = ["<table>"]
        data.append("<tr><th></th><th>%s</th></tr>" %
                    "</th><th>".join(map(str, lines[0].split(","))))
        data.extend(["<tr><td>%s</td></tr>" %
                     ("</td><td>".join(x.split(","))) for x in lines[1:]])
        data.append("</table>\n")

        # substitute links
        data = [re.sub(
            "`(.*?(?:\".+\"|\'.+\')?.*?)\s<(.*?(?:\".+\"|\'.+\')?.*?)>`_",
            r'<a href="\2">\1</a>',
            x)
            for x in data]

        r.html = "\n".join(data)

        return r

    def asSpreadSheet(self, dataframe, row_headers, col_headers, title):
        '''save the table as an xls file.

        Multiple files of the same Renderer/Tracker combination are
        distinguished by the title.
        '''

        self.debug("%s: saving %i x %i table as spread-sheet'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))

        is_hierarchical = isinstance(dataframe.index,
                                     pandas.core.index.MultiIndex)

        split = is_hierarchical and len(dataframe.index.levels) > 1

        quick = len(dataframe) > 10000
        if quick and not split:
            # quick writing, only append method works
            wb = openpyxl.Workbook()

            def fillWorksheet(ws, dataframe, title):
                ws.append([""] + list(col_headers))
                for x, row in enumerate(dataframe.iterrows()):
                    ws.append([path2str(row[0])] + list(row[1]))

                # patch: maximum title length seems to be 31
                ws.title = title[:30]

        else:
            # do it cell-by-cell, this might be slow
            wb = openpyxl.Workbook()

            def fillWorksheet(ws, dataframe, title):
                # regex to detect rst hypelinks
                regex_link = re.compile('`(.*) <(.*)>`_')
                # write row names
                for row, row_name in enumerate(dataframe.index):
                    # rows and columns start at 1
                    c = ws.cell(row=row + 2, column=1)
                    c.value = row_name

                # write columns
                for column, column_name in enumerate(dataframe.columns):
                    # set column title
                    # rows and columns start at 1
                    c = ws.cell(row=1, column=column + 2)
                    c.value = column_name

                    # set column values
                    dataseries = dataframe[column_name]

                    if dataseries.dtype == object:
                        for row, value in enumerate(dataseries):
                            c = ws.cell(row=row + 2,
                                        column=column + 2)
                            value = str(value)
                            if value.startswith('`'):
                                c.value, c.hyperlink =\
                                    regex_link.match(value).groups()
                            else:
                                c.value = value
                    else:
                        for row, value in enumerate(dataseries):
                            c = ws.cell(row=row + 2,
                                        column=column + 2)
                            c.value = value
                if title:
                    # patch: maximum title length seems to be 31
                    ws.title = re.sub("/", "_", title)[:30]

        if len(wb.worksheets) == 0:
            wb.create_sheet()

        if split:
            # create separate worksheets for nested indices
            nlevels = len(dataframe.index.levels)
            paths = list(map(tuple, DataTree.unique(
                [x[:nlevels - 1]
                 for x in dataframe.index.unique()])))

            ws = wb.worksheets[0]
            ws.title = 'Summary'
            ws.append(
                [""] * (nlevels - 1) + ["Worksheet", "Rows"])

            for row, path in enumerate(paths):
                # select data frame as cross-section
                work = dataframe.xs(path, axis=0)
                title = path2str(path)
                if len(title) > 30:
                    title = "sheet%i" % row

                ws.append(list(path) + [title, len(work)])
                c = ws.cell(row=row + 1,
                            column=nlevels)
                # this does not work in oocalc
                c.hyperlink = "#%s!A1" % title
                fillWorksheet(wb.create_sheet(),
                              work,
                              title=title)
        else:
            fillWorksheet(wb.worksheets[0], dataframe,
                          title=title)

        # write result block
        lines = []
        lines.append("`%i x %i table <#$xls %s$#>`__" %
                     (len(row_headers), len(col_headers),
                      title))
        lines.append("")

        r = ResultBlock("\n".join(lines), title=title)
        r.xls = wb

        self.debug("%s: saved %i x %i table as spread-sheet'" %
                   (id(self),
                    len(row_headers),
                    len(col_headers)))
        return r


class Table(TableBase):

    '''a basic table.

    Values are either text or converted to text.

    If the has more rows than:attr:`max_rows` or more columns
    than:attr:`max_cols`, a placeholder will be inserted pointing
    towards a file.

    The attribute:attr:`large` determines where large tables are
    written to. The default is html. Alternative values are ``xls``
    for excel spread-sheets.

    '''
    options = TableBase.options +\
        (('transpose', directives.unchanged),
         ('add-percent', directives.unchanged),
         ('head', directives.length_or_unitless),
         ('large', directives.unchanged),
         ('format-columns', directives.unchanged),
         ('preview', directives.unchanged))

    # By default, group everything together
    group_level = "all"

    transpose = False

    def __init__(self, *args, **kwargs):
        TableBase.__init__(self, *args, **kwargs)

        self.transpose = "transpose" in kwargs
        self.large = kwargs.get("large", "html")
        self.preview = "preview" in kwargs

        if "format-columns" in kwargs:
            self.format_columns = kwargs.get("format-columns",
                                             None)
            if not self.format_columns:
                self.format_columns = "auto"
            else:
                columns = self.format_columns.split(",")
                self.format_columns = {}
                for column in columns:
                    if "=" in column:
                        column, fmt = column.split("=")
                    else:
                        fmt = "auto"
                    self.format_columns[column] = fmt
        else:
            self.format_columns = None

        self.add_percent = kwargs.get('add-percent', None)
        self.head = int(kwargs.get('head', 0))

    def modifyTable(self, dataframe):
        '''modify table if required, for example adding percentages.
        '''

        if self.head > 0:
            dataframe = dataframe[:self.head]

        if self.add_percent:
            columns = dataframe.columns
            parts = self.add_percent.split(";")
            for part in parts:
                total, other_col = None, None
                if "," in part:
                    column, method = part.split(",")
                    if method in columns:
                        other_col = columns.index(method)
                    else:
                        try:
                            total = float(method)
                        except ValueError:
                            raise ValueError(
                                ("can't compute total from "
                                 "expression `%s` - missing column?") %
                                method)
                else:
                    column = part

                if column not in columns:
                    raise ValueError("unknown column `%s` to add-percent" %
                                     (column))

                values = numpy.array(dataframe.sum(axis=1),
                                     dtype=numpy.float)

                if other_col:
                    dataframe['%s/%%' % column] = 100.0 *\
                        dataframe[column] / dataframe[other_col]
                else:
                    if total is None:
                        total = float(sum(values))
                    dataframe['%s/%%' % column] = 100.0 *\
                        dataframe[column] / values

        if self.transpose:
            dataframe = dataframe.transpose()
            # flatten the column index if it is hierarchical
            is_hierarchical = isinstance(dataframe.columns,
                                         pandas.core.index.MultiIndex)
            if is_hierarchical:
                dataframe.columns = ["/".join(x) for x in dataframe.columns]

        if self.format_columns is not None:
            for column in dataframe.columns:
                if self.format_columns == "auto":
                    fmt = "auto"
                elif column in self.format_columns:
                    fmt = self.format_columns[column]
                else:
                    continue

                cc = dataframe[column]
                if cc.dtype == numpy.float:
                    mi, ma = cc.min(), cc.max()
                    if mi >= 0 and ma <= 1.0:
                        # convert to percent
                        dataframe[column] = (100.0 * cc).map('{:,.2f}'.format)
                    else:
                        dataframe[column] = cc.map('{:,.2f}'.format)
                elif cc.dtype == numpy.int:
                    dataframe[column] = cc.map('{:,}'.format)

        # if index is not hierarchical, but contains tuples,
        # split tuples in index to build a new (hierarchical) index
        is_hierarchical = isinstance(dataframe.index,
                                     pandas.core.index.MultiIndex)

        if not is_hierarchical and isinstance(dataframe.index[0], tuple):
            idx = pandas.MultiIndex.from_tuples(dataframe.index)
            dataframe.index = idx

        return dataframe

    def __call__(self, dataframe, path):

        # modify table (adding/removing columns) according to user options
        # matrix, row_headers, col_headers = \
        # self.modifyTable(matrix, row_headers, col_headers)
        dataframe = self.modifyTable(dataframe)

        title = path2str(path)

        results = ResultBlocks()

        row_headers = dataframe.index
        col_headers = dataframe.columns

        # as of sphinx 1.3.1, tables with more than 100 columns cause an
        # error:
        # Exception occurred:
        # File "/ifs/apps/apps/python-2.7.9/lib/python2.7/site-packages/docutils/
        # writers/html4css1/__init__.py", line 642, in write_colspecs
        # colwidth = int(node['colwidth'] * 100.0 / width + 0.5)
        # ZeroDivisionError: float division by zero
        #
        # Thus, for table with more than 100 columns, force will be
        # disabled and max_cols set to a low value in order to make
        # sure the table is not displayed inline
        if len(col_headers) >= 90:
            self.force = False
            self.max_cols = 10

        # do not output large matrices as rst files
        if self.separate or (not self.force and
                             (len(row_headers) > self.max_rows or
                              len(col_headers) > self.max_cols)):
            if self.large == "xls":
                results.append(self.asSpreadSheet(dataframe, row_headers,
                                                  col_headers, title))
            else:
                results.append(self.asFile(dataframe, row_headers,
                                           col_headers, title))

            if self.preview:
                raise NotImplementedError('preview not implemented')
                row_headers = row_headers[:self.max_rows]
                col_headers = col_headers[:self.max_cols]
                # matrix = [x[:self.max_cols] for x in
                #          matrix[:self.max_rows]]
            else:
                return results

        results.append(self.asCSV(dataframe, row_headers, col_headers, title))

        return results


class HTMLTable(Table):

    '''a table in html format to download.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()
        if dataframe is None:
            return results

        title = path2str(path)
        row_headers = dataframe.index
        col_headers = dataframe.columns
        results = ResultBlocks()
        results.append(self.asFile(dataframe, row_headers,
                                   col_headers, title))

        return results


class XlsTable(Table):

    '''a table in xls format to download.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()
        if dataframe is None:
            return results

        title = path2str(path)
        row_headers = dataframe.index
        col_headers = dataframe.columns
        results.append(self.asSpreadSheet(dataframe, row_headers,
                                          col_headers, title))

        return results


class RstTable(Table):

    '''an rst formatted table.
    '''

    def __call__(self, dataframe, path):

        results = ResultBlocks()

        if dataframe is None:
            return results

        title = path2str(path)

        row_headers = dataframe.index
        col_headers = dataframe.columns

        # do not output large matrices as rst files
        if self.separate or (not self.force and
                             (len(row_headers) > self.max_rows or
                              len(col_headers) > self.max_cols)):
            if self.large == "xls":
                results.append(self.asSpreadSheet(dataframe, row_headers,
                                                  col_headers, title))
            else:
                results.append(self.asFile(dataframe, row_headers,
                                           col_headers, title))

        results.append(self.asRST(dataframe, row_headers, col_headers, title))

        return results


class GlossaryTable(Table):

    """output a table in the form of a glossary."""

    def __call__(self, dataframe, path):

        results = ResultBlocks()

        if dataframe is None:
            return results

        title = path2str(path)

        lines = []
        lines.append(".. glossary::")
        lines.append("")

        for x, row in enumerate(dataframe.iterrows()):
            header, data = row
            txt = "\n      ".join([x.strip() for x in str(data).split("\n")])
            lines.append('   %s\n      %s\n' % (path2str(header), txt))

        lines.append("")

        results.append(ResultBlock("\n".join(lines), title=title))

        return results


class MatrixBase:
    '''base class for matrices.

    This base class provides utility functions for rectangular 2D matrices.

    It implements column-wise and row-wise transformations.

    This class adds the following options to the:term:`report` directive.

    :term:`transform-matrix`: apply a matrix transform. Possible choices are:

           * *correspondence-analysis*: use correspondence analysis
             to permute rows/columns
           * *normalized-col-total*: normalize by column total
           * *normalized-col-max*: normalize by column maximum,
           * *normalized-col-first*: normalize by first column.
             The first column is then removed.
           * *normalized-row-total*: normalize by row total
           * *normalized-row-max*: normalize by row maximum
           * *normalized-row-first*: normalize by first row.
             The first row is then removed.
           * *normalized-total*: normalize over whole matrix
           * *normalized-max*: normalize over whole matrix
           * *sort*: sort matrix rows and columns alphanumerically.
           * *sort-numerically*: sort matrix rows and columns numerically.
           * filter-by-rows: only take columns that are also present in rows
           * filter-by-cols: only take columns that are also present in cols
           * square: make square matrix (only take rows and columns present
             in both)
           * *add-row-total*: add the row total at the bottom
           * *add-column-total*: add the column total as a last column

    Requires two levels:

       rows[dict] / columns[dict] / value

    All values need to be numerical.
    '''

    def __init__(self, *args, **kwargs):

        self.mMapKeywordToTransform = {
            "correspondence-analysis": self.transformCorrespondenceAnalysis,
            "transpose": self.transformTranspose,
            "normalized-row-total": self.transformNormalizeRowTotal,
            "normalized-row-max": self.transformNormalizeRowMax,
            "normalized-row-first": self.transformNormalizeRowFirst,
            "normalized-col-total": self.transformNormalizeColumnTotal,
            "normalized-col-max": self.transformNormalizeColumnMax,
            "normalized-col-first": self.transformNormalizeColumnFirst,
            "normalized-total": self.transformNormalizeTotal,
            "normalized-max": self.transformNormalizeMax,
            "symmetric-max": self.transformSymmetricMax,
            "symmetric-min": self.transformSymmetricMin,
            "symmetric-avg": self.transformSymmetricAverage,
            "symmetric-sum": self.transformSymmetricSum,
            "filter-by-rows": self.transformFilterByRows,
            "filter-by-cols": self.transformFilterByColumns,
            "square": self.transformSquare,
            "add-row-total": self.transformAddRowTotal,
            "add-column-total": self.transformAddColumnTotal,
            "sort": self.transformSort,
            "sort-numerically": self.transformSortNumerically,
        }
        self.tofloat = False
        self.converters = []
        if "transform-matrix" in kwargs:
            for kw in [x.strip() for x in
                       kwargs["transform-matrix"].split(",")]:
                if kw.startswith("normalized"):
                    self.format = "%6.4f"
                    self.tofloat = True
                try:
                    self.converters.append(self.mMapKeywordToTransform[kw])
                except KeyError:
                    raise ValueError(("unknown matrix transformation %s, "
                                      "possible values are: %s") %
                                     (kw,
                                      ",".join(sorted(
                                          self.mMapKeywordToTransform.keys()
                                      ))))

    def transformAddRowTotal(self, matrix, row_headers, col_headers):
        '''add row total to the matrix.'''
        nrows, ncols = matrix.shape

        totals = numpy.zeros(nrows)
        for x in range(nrows):
            totals[x] = sum(matrix[x, :])
        col_headers.append("total")
        totals.shape = (nrows, 1)
        return (numpy.hstack(numpy.hsplit(matrix, ncols) + [totals, ]),
                row_headers, col_headers)

    def transformAddColumnTotal(self, matrix, row_headers, col_headers):
        raise NotImplementedError

    def transformFilterByRows(self, matrix, row_headers, col_headers):
        """only take columns that are also present in rows"""
        take = [x for x in range(len(col_headers))
                if col_headers[x] in row_headers]
        return (matrix.take(take, axis=1), row_headers,
                [col_headers[x] for x in take])

    def transformFilterByColumns(self, matrix, row_headers, col_headers):
        """only take rows that are also present in columns"""
        take = [x for x in range(len(row_headers))
                if row_headers[x] in col_headers]
        return (matrix.take(take, axis=0),
                [row_headers[x] for x in take],
                col_headers)

    def transformSquare(self, matrix, row_headers, col_headers):
        """only take rows and columns that are present in both giving a square
        matrix.
        """
        take = set(row_headers).intersection(set(col_headers))
        row_indices = [x for x in range(len(row_headers))
                       if row_headers[x] in take]
        col_indices = [x for x in range(len(col_headers))
                       if col_headers[x] in take]
        m1 = matrix.take(row_indices, axis=0)
        m2 = m1.take(col_indices, axis=1)

        return (m2,
                [row_headers[x] for x in row_indices],
                [col_headers[x] for x in col_indices])

    def transformTranspose(self, matrix, row_headers, col_headers):
        """transpose the matrix."""
        return numpy.transpose(matrix), col_headers, row_headers

    def transformCorrespondenceAnalysis(self, matrix,
                                        row_headers,
                                        col_headers):
        """apply correspondence analysis to a matrix.
        """
        if len(row_headers) <= 1 or len(col_headers) <= 1:
            self.warn("correspondence analysis skipped for "
                      "matrices with single row/column")
            return matrix, row_headers, col_headers

        try:
            row_indices, col_indices = \
                CorrespondenceAnalysis.GetIndices(matrix)
        except ValueError:
            return matrix, row_headers, col_headers

        map_row_new2old = numpy.argsort(row_indices)
        map_col_new2old = numpy.argsort(col_indices)

        matrix, row_headers, col_headers = \
            CorrespondenceAnalysis.GetPermutatedMatrix(matrix,
                                                       map_row_new2old,
                                                       map_col_new2old,
                                                       row_headers=row_headers,
                                                       col_headers=col_headers)

        return matrix, row_headers, col_headers

    def transformSort(self, matrix, row_headers, col_headers):
        """apply correspondence analysis to a matrix.
        """

        map_row_new2old = [x[0] for x in sorted(enumerate(row_headers),
                                                key=lambda x: x[1])]
        map_col_new2old = [x[0] for x in sorted(enumerate(col_headers),
                                                key=lambda x: x[1])]

        matrix, row_headers, col_headers = \
            CorrespondenceAnalysis.GetPermutatedMatrix(matrix,
                                                       map_row_new2old,
                                                       map_col_new2old,
                                                       row_headers=row_headers,
                                                       col_headers=col_headers)

        return matrix, row_headers, col_headers

    def transformSortNumerically(self, matrix, row_headers, col_headers):
        """apply correspondence analysis to a matrix.
        """

        def atoi(text):
            return int(text) if text.isdigit() else text

        def natural_keys(text):
            '''
            alist.sort(key=natural_keys) sorts in human order
            http://nedbatchelder.com/blog/200712/human_sorting.html
            (See Toothy's implementation in the comments)
            '''
            return [atoi(c) for c in re.split('(\d+)', text)]

        map_row_new2old = [x[0] for x in sorted(
            enumerate([natural_keys(x) for x in row_headers]),
            key=lambda x: x[1])]

        map_col_new2old = [x[0] for x in sorted(
            enumerate([natural_keys(x) for x in col_headers]),
            key=lambda x: x[1])]

        matrix, row_headers, col_headers = \
            CorrespondenceAnalysis.GetPermutatedMatrix(matrix,
                                                       map_row_new2old,
                                                       map_col_new2old,
                                                       row_headers=row_headers,
                                                       col_headers=col_headers)

        return matrix, row_headers, col_headers

    def transformSymmetricMax(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = max(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricMin(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    min(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricSum(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    sum(matrix[x, y], matrix[y, x])

        return matrix, rows, cols

    def transformSymmetricAverage(self, matrix, rows, cols):
        """symmetrize a matrix.

        returns the normalized matrix.
        """
        if len(rows) != len(cols):
            raise ValueError("matrix not square - can not be symmetrized")

        for x in range(len(rows)):
            for y in range(x + 1, len(cols)):
                matrix[x, y] = matrix[y, x] = \
                    sum(matrix[x, y], matrix[y, x]) / 2

        return matrix, rows, cols

    def transformNormalizeTotal(self, matrix, rows, cols):
        """normalize a matrix by the total.

        Returns the normalized matrix.
        """
        return matrix / matrix.sum(), rows, cols

    def transformNormalizeMax(self, matrix, rows, cols):
        """normalize a matrix by the max.

        Returns the normalized matrix.
        """
        return matrix / matrix.max(), rows, cols

    def transformNormalizeRowTotal(self, matrix, rows, cols):
        """normalize a matrix row by the row total.

        Returns the normalized matrix.
        """
        return matrix / matrix.sum(axis=1), rows, cols

    def transformNormalizeRowMax(self, matrix, rows, cols):
        """normalize a matrix row by the row maximum.

        Returns the normalized matrix.
        """
        return matrix / matrix.max(axis=1), rows, cols

    def transformNormalizeRowFirst(self, matrix, rows, cols):
        """normalize a matrix row by the row maximum.

        Returns the normalized matrix.
        """
        m = matrix[0]
        matrix = numpy.delete(matrix, 0, 0)
        return matrix / m, rows[1:], cols

    def transformNormalizeColumnTotal(self, matrix, rows, cols):
        """normalize a matrix by the column total.

        Returns the normalized matrix."""
        return matrix / matrix.sum(axis=0), rows, cols

    def transformNormalizeColumnFirst(self, matrix, rows, cols):
        """normalize a matrix by the first column.

        Removes the first column.

        Returns the normalized matrix."""
        m = matrix[:, 0]
        matrix = numpy.delete(matrix, 0, 1)
        return matrix / m, rows, cols[1:]

    def transformNormalizeColumnMax(self, matrix, rows, cols):
        """normalize a matrix by the column maximum

        Returns the normalized matrix.
        """
        return matrix / matrix.max(axis=0), rows, cols

    def render(self, work, path):
        """render the data.
        """

        results = ResultBlocks(title=path)

        matrix, rows, columns = self.buildMatrix(work)

        title = path2str(path)

        if len(rows) == 0:
            return ResultBlocks(ResultBlock("", title=title))

        # do not output large matrices as rst files
        # separate and force need to be mixed in.
        if self.separate or (not self.force and
                             (len(rows) > self.max_rows or
                              len(columns) > self.max_cols)):
            return ResultBlocks(self.asFile(pandas.DataFrame(matrix,
                                                             index=rows,
                                                             columns=columns),
                                            rows,
                                            columns,
                                            title),
                                title=path)

        lines = []
        lines.append(".. csv-table:: %s" % title)
        lines.append('   :header: "track","%s" ' % '","'.join(columns))
        lines.append('')
        for row in range(len(rows)):
            lines.append(
                '   "%s","%s"' %
                (rows[row], '","'.join(
                    [self.toString(x) for x in matrix[row]])))
        lines.append("")

        if path is None:
            subtitle = ""
        else:
            subtitle = path2str(path)

        results.append(ResultBlock("\n".join(lines), title=subtitle))

        return results


class TableMatrix(TableBase, MatrixBase):

    """A table with numerical columns.

       rows[dict] / columns[dict] / value

    All values need to be numerical.
    """

    # By default, group everything together
    group_level = "all"

    options = TableBase.options +\
        (('full-row-labels', directives.unchanged),
         ('transform-matrix', directives.unchanged),)

    # wether or not to convert matrix to float
    tofloat = False

    def __init__(self, *args, **kwargs):

        TableBase.__init__(self, *args, **kwargs)
        MatrixBase.__init__(self, *args, **kwargs)

        self.normalize_row_labels = "full-row-labels" not in kwargs

    def buildMatrix(self,
                    dataframe,
                    missing_value=0,
                    apply_transformations=True,
                    dtype=numpy.float):
        """build a matrix from work, a two-level nested dictionary.

        This method will also apply conversions if apply_transformations
        is set.
        """

        if self.normalize_row_labels:
            drop = [x for x, y in enumerate(dataframe.index.values)
                    if len(set(y)) == 1]
            rows = list(map(path2str, dataframe.index.droplevel(drop)))
        else:
            rows = list(map(path2str, dataframe.index))

        columns = list(dataframe.columns)
        # use numpy.matrix - permits easier broadcasting
        # for normalization.
        matrix = numpy.matrix(dataframe.as_matrix())

        # remove columns with only NaNs. This can happend
        # during the dataframe merging process if the
        # columns are unique to each matrix.
        take = numpy.array(numpy.all(numpy.isnan(matrix), axis=0).flat)
        matrix = matrix[:, ~take]
        columns = list(numpy.array(columns)[~take])

        if self.converters and apply_transformations:
            # convert to float for conversions
            if self.tofloat:
                matrix = numpy.matrix(matrix, dtype=numpy.float)

            for converter in self.converters:
                self.debug("applying converter %s" % converter)
                matrix, rows, columns = converter(matrix, rows, columns)

        # convert rows/columns to str (might be None)
        rows = [str(x) for x in rows]
        columns = [str(x) for x in columns]

        # convert numpy.matrix to numpy.array
        return numpy.asarray(matrix), rows, columns

# for compatibility
Matrix = TableMatrix


class NumpyMatrix(TableMatrix, MatrixBase):

    """Deprecated - not needed any more as equivalent to TableMatrix
    """

    def __init__(self, *args, **kwargs):
        TableMatrix.__init__(self, *args, **kwargs)
        MatrixBase.__init__(self, *args, **kwargs)


class Status(Renderer):
    '''Renders a status report.

    A status report is a two element table containing
    status ('PASS', 'FAIL', 'WARNING', 'NA') and some information.

    The __doc__ string of the tracker is added as a legend below
    the status report.
    '''

    options = Renderer.options +\
        (('columns', directives.unchanged),
         ('no-legend', directives.unchanged),)

    # read complete data
    nlevels = -1

    # For backwards compatibility, group by slice (test function)
    # and not track (data track) by default.
    group_level = 'slice'

    map_code2image = {'FAIL': "fail.png",
                      'PASS': "pass.png",
                      'NA': "not_available.png",
                      'WARNING': "warning.png",
                      'WARN': "warning.png"}

    display_legend = True

    columns = ["track", "test", "image", "status", "info"]

    def __init__(self, *args, **kwargs):

        Renderer.__init__(self, *args, **kwargs)
        if "no-legend" in kwargs:
            self.display_legend = False
        if "columns" in kwargs:
            self.columns = [x.strip() for x in kwargs["columns"].split(",")]

    def __call__(self, dataframe, path):

        # convert to dataframe
        # index has test names
        # columns are description, info, status
        columns = ('description', 'info', 'status', 'name')
        if set(dataframe.columns) != set(columns):
            raise ValueError("invalid columns: expected '%s', got '%s' " %
                             (columns, dataframe.columns))

        lines = []
        dirname = os.path.join(os.path.dirname(
            sys.modules["CGATReport"].__file__), "images")
        descriptions = {}
        title = "status"

        # add header
        lines.append(".. csv-table:: %s" % "table")
        lines.append('   :header: "Track", "Test", "", "Status", "Info"')
        lines.append('')
        rows = []
        for index, values in dataframe.iterrows():
            testname = values['name']
            status = values['status']
            try:
                image = ".. image:: {}\n    :width: 32".format(
                    os.path.join(dirname,
                                 self.map_code2image[status.upper()]))
            except KeyError:
                image = ""

            rows.append({
                "test": testname,
                "description": values["description"],
                "info": values['info'],
                "status": status,
                "track": path2str(index),
                "image": image,
            })
            descriptions[testname] = values["description"]

        # filter and sort table
        table = [self.columns]
        table.extend([[row[x] for x in self.columns] for row in rows])

        lines = Utils.table2rst(table).split("\n")

        if self.display_legend:
            lines.append(".. glossary::")
            lines.append("")

            for test, description in list(descriptions.items()):
                lines.append('%s\n%s\n' % (Utils.indent(test, 3),
                                           Utils.indent(description, 6)))

        return ResultBlocks(ResultBlock("\n".join(lines), title=""))


class StatusMatrix(Status, TableBase):
    '''Renders a status matrix.

    A status matrix is a rectangular matrix displaying icons for
    status flags ('PASS', 'FAIL', 'WARNING', 'NA').

    '''
    options = Status.options + TableBase.options +\
        (('row-column', directives.unchanged),
         ('transpose', directives.unchanged),)

    # For backwards compatibility, group by slice (test function)
    # and not track (data track) by default.
    group_level = None

    transpose = False

    row_column = "track"

    def __init__(self, *args, **kwargs):
        TableBase.__init__(self, *args, **kwargs)
        Status.__init__(self, *args, **kwargs)

        self.transpose = "transpose" in kwargs
        self.row_column = kwargs.get("row-column", "track")

    def __call__(self, dataframe, path):

        dirname = os.path.join(os.path.dirname(
            sys.modules["CGATReport"].__file__), "images")

        # modify columns in-place, don't do this with original slice
        dataframe = dataframe.copy()

        dataframe["status"] = [
            ".. image:: {}\n    :width: 32".format(
                os.path.join(dirname,
                             self.map_code2image.get(x.upper(), "NA")))
            for x in dataframe["status"]]

        table = pandas.pivot_table(
            dataframe.reset_index(),
            index=[self.row_column],
            columns="slice",
            values="status",
            aggfunc=lambda x: ' '.join(x))

        if self.transpose:
            table = table.transpose()
            # flatten the column index if it is hierarchical
            is_hierarchical = isinstance(table.columns,
                                         pandas.core.index.MultiIndex)
            if is_hierarchical:
                table.columns = ["/".join(x) for x in table.columns]

        results = ResultBlocks()
        results.append(self.asCSV(table, [], [], path2str(path)))
        return results


class DataTreeRenderer(object):
    """Base class for renderers processing a data tree
    directly.

    Note that these are special case renderers.
    """
    # plugin fields
    capabilities = ["render"]

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, data):
        return self.render(data)


class User(DataTreeRenderer):

    """Renderer for user-implemented rendering.

    The renderer itself creates no output, but returns the results of
    the tracker.

    When called, a Renderer and its subclasses will return blocks of
    restructured text. Images are automatically collected from
    matplotlib and other renderers from active graphics devices and
    inserted at the place-holders.

    """

    def __init__(self, *args, **kwargs):
        DataTreeRenderer.__init__(self, *args, **kwargs)

    def render(self, data):

        # initiate output structure
        results = ResultBlocks(title='user')

        labels = DataTree.getPaths(data)
        # iterate over all items at leaf
        for path, branch in DataTree.getNodes(data, len(labels) - 2):
            for key in Utils.TrackerKeywords:
                if key in branch:
                    # add a result block
                    results.append(ResultBlock(branch[key],
                                               title=path2str(path)))

        return results


class Debug(DataTreeRenderer):

    '''a simple renderer, returning the type of data
    and the number of items at each path.'''

    def render(self, data):

        # initiate output structure
        results = ResultBlocks(title='debug')

        try:
            results.append(ResultBlock(json.dumps(
                data, indent=4), title=''))
        except TypeError:
            results.append(ResultBlock(str(data), title=''))

        return results



#!/bin/env python

"""
cgatreport-get
================

:command:`cgatreport-get` queries the CGATReport cache
and retrieves data.

   cgatreport-get tracker

The full list of command line options is listed by suppling:option:`-h/--help`
on the command line.

The options are:

**-t/--tracker** tracker
:class:`Tracker` to use.

**-a/--tracks** tracks
   Tracks to display as a comma-separated list. If none are given, output all tracks.

**-s/--slices** slices
   Slices to display as a comma-separated list. If none are given, output all slices

**-v/--view**
   Do not ouput data, but display list of available tracks and slices.

**-g/--groupby** group by
   (track,slice)
   Group output either by:term:`track` or:term:`slice`.

**-f/--format** output format
   (tsv, csv)
   Output format. Available are:

   * ``tsv``: tab-separated values
   * ``csv``: comma-separated values

"""
import sys
import os
import imp
import io
import re
import types
import glob
import optparse
import shutil

USAGE = """python %s [OPTIONS]

query the CGATReport cache.

""" % sys.argv[0]


from CGATReport import Utils
from CGATReport import Cache
from CGATReport import DataTree
from collections import OrderedDict as odict


def main():

    parser = optparse.OptionParser(version="%prog version: $Id$", usage=USAGE)

    parser.add_option("-v", "--verbose", dest="loglevel", type="int",
                      help="loglevel. The higher, the more output [default=%default]")

    parser.add_option("-i", "--view", dest="view", action="store_true",
                      help="view keys in cache [default=%default]")

    parser.add_option("-t", "--tracker", dest="tracker", type="string",
                      help="tracker to use [default=%default]")

    parser.add_option("-a", "--tracks", dest="tracks", type="string",
                      help="tracks to include [default=%default]")

    parser.add_option("-s", "--slices", dest="slices", type="string",
                      help="slices to include [default=%default]")

    parser.add_option("-g", "--groupby", dest="groupby", type="choice",
                      choices=("track", "slice", "all"),
                      help="groupby by track or slice [default=%default]")

    parser.add_option("-f", "--format", dest="format", type="choice",
                      choices=("tsv", "csv"),
                      help="output format [default=%default]")

    parser.set_defaults(
        loglevel=2,
        view=False,
        tracker=None,
        tracks=None,
        slices=None,
        groupby="slice",
        format="tsv",
    )

    (options, args) = parser.parse_args()

    if len(args) != 1 and options.tracker == None:
        print(USAGE)
        raise ValueError("please supply a tracker.""")

    if options.tracker:
        tracker = options.tracker
    else:
        tracker = args[0]

    cache = Cache.Cache(tracker, mode="r")

    if options.view:
        keys = [x.split("/") for x in list(cache.keys())]
        sys.stdout.write("# available tracks\n")
        sys.stdout.write("track\n%s" % "\n".join(set([x[0] for x in keys])))
        sys.stdout.write("\n")
        sys.stdout.write("# available slices\n")
        sys.stdout.write("slice\n%s" % "\n".join(set([x[1] for x in keys])))
        sys.stdout.write("\n")
        return

    data = DataTree.fromCache(cache,
                              tracks=options.tracks,
                              slices=options.slices,
                              groupby=options.groupby)

    table, row_headers, col_headers = DataTree.tree2table(data)

    if options.format in ("tsv", "csv"):
        if options.format == "tsv":
            sep = "\t"
        elif options.format == "csv":
            sep = ","
        sys.stdout.write(sep + sep.join(col_headers) + "\n")
        for h, row in zip(row_headers, table):
            sys.stdout.write("%s%s%s\n" % (h, sep, sep.join(row)))

if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
"""
    sphinx.ext.errors
    ~~~~~~~~~~~~~~~~~

    Collect all errors

:copyright: Copyright 2007-2009 by the Andreas Heger, see AUTHORS.
:license: BSD, see LICENSE for details.
"""

from docutils import nodes
import os
import logging

from sphinx.locale import _
from sphinx.environment import NoUri
from sphinx.util.compat import Directive
from docutils.parsers.rst.directives.admonitions import BaseAdmonition
import shelve

CGATREPORT_ERRORS_CACHE = "cgatreport_errors.cache"


class cgatreporterror_node(nodes.warning, nodes.Element):
    pass


class cgatreporterrorlist(nodes.General, nodes.Element):
    pass


class CGATReportError(BaseAdmonition):

    node_class = cgatreporterror_node

    # accept error as an optional argument
    optional_arguments = 1

    def run(self):

        r = BaseAdmonition.run(self)

        if len(self.arguments) > 0:
            errorclass = self.arguments[0]
        else:
            errorclass = "generic"

        env = self.state.document.settings.env

        error_cache = shelve.open(CGATREPORT_ERRORS_CACHE)

        error_cache["{}:{:06}".format(env.docname, self.lineno)] = (
            r[0].deepcopy(), errorclass)

        logging.error("CGATReport-Error: %s" % errorclass)

        return r


class CGATReportErrorList(Directive):

    """
    A list of all cgatreporterror entries.
    """

    has_content = False
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {}

    def run(self):
        # Simply insert an empty cgatreporterrorlist node which will
        # be replaced later when process_cgatreporterror_nodes is
        # called
        return [cgatreporterrorlist('')]


def process_cgatreporterror_nodes(app, doctree, fromdocname):

    if not app.config['cgatreport_show_errors']:
        for node in doctree.traverse(cgatreporterror_node):
            node.parent.remove(node)

    # Replace all cgatreporterrorlist nodes with a list of the
    # collected cgatreporterrors.  Augment each cgatreporterror with a
    # backlink to the original location.
    env = app.builder.env

    error_cache = shelve.open(CGATREPORT_ERRORS_CACHE)

    for node in doctree.traverse(cgatreporterrorlist):

        if not app.config['cgatreport_show_errors']:
            node.replace_self([])
            continue

        nerrors = 0

        content = []

        para = nodes.paragraph()
        sorted_items = sorted(error_cache.items())
        para += nodes.Text("There are {} errors".format(len(sorted_items)))

        content.append(para)

        for key, value in sorted_items:

            docname, lineno = key.split(":")
            lineno = int(lineno)
            cgatreporterror, errorclass = value

            para = nodes.paragraph()

            filename = env.doc2path(docname, base=None)

            nerrors += 1
            location_str = '%s:%d ' % (filename, lineno)

            try:
                description_str = errorclass
            except KeyError:
                description_str = "unknown"

            # Create a reference
            newnode = nodes.reference('', '')
            innernode = nodes.emphasis((location_str), (location_str))
            newnode['refdocname'] = docname
            try:
                newnode['refuri'] = app.builder.get_relative_uri(
                    fromdocname, docname)
            except NoUri:
                # ignore if no URI can be determined, e.g. for LaTeX output
                pass

            newnode.append(innernode)

            para += newnode
            para += nodes.Text(description_str, description_str)
            para += nodes.Text("\n", "\n")

            # (Recursively) resolve references in the cgatreporterror content
            env.resolve_references(cgatreporterror,
                                   docname,
                                   app.builder)

            content.append(para)

        node.replace_self(content)


def purge_cgatreporterrors(app, env, docname):
    if not hasattr(env, 'cgatreporterror_all_cgatreporterrors'):
        return
    env.cgatreporterror_all_cgatreporterrors = \
        [cgatreporterror for cgatreporterror in
         env.cgatreporterror_all_cgatreporterrors
         if cgatreporterror['docname'] != docname]


def visit_cgatreporterror_node(self, node):
    self.visit_admonition(node)


def depart_cgatreporterror_node(self, node):
    self.depart_admonition(node)


def setup(app):
    app.add_config_value('cgatreport_show_errors', True, False)

    app.add_node(cgatreporterrorlist)
    app.add_node(
        cgatreporterror_node,
        html=(
            visit_cgatreporterror_node, depart_cgatreporterror_node),
        latex=(
            visit_cgatreporterror_node,
            depart_cgatreporterror_node),
        text=(visit_cgatreporterror_node, depart_cgatreporterror_node))

    app.add_directive('error', CGATReportError)
    app.add_directive('errorlist', CGATReportErrorList)
    app.connect('doctree-resolved', process_cgatreporterror_nodes)
    app.connect('env-purge-doc', purge_cgatreporterrors)

    try:
        os.unlink(CGATREPORT_ERRORS_CACHE)
    except OSError:
        pass

    return {'parallel_read_safe': True}
